#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import io
import pathlib
import subprocess

import pytest
from craft_providers.multipass import MultipassInstance

from . import conftest

# These tests can be flaky on a sufficiently busy system, because multipass will
# sometime fail to talk to the VM. If they fail, retry them after a short delay.
pytestmark = pytest.mark.flaky(reruns=3, reruns_delay=2)


@pytest.fixture()
def instance(instance_name):
    with conftest.tmp_instance(
        instance_name=instance_name,
    ) as tmp_instance:
        yield MultipassInstance(name=tmp_instance)


@pytest.fixture(scope="module")
def reusable_instance(reusable_instance_name):
    """Reusable instance for tests that don't require a fresh instance."""
    with conftest.tmp_instance(
        instance_name=reusable_instance_name,
    ) as tmp_instance:
        yield MultipassInstance(name=tmp_instance)


@pytest.fixture()
def simple_file(home_tmp_path):
    """Create a file in the home directory (accessible by Multipass)."""
    file = home_tmp_path / "src.txt"
    file.write_text("this is a test")
    return file


@pytest.mark.parametrize("content", [b"", b"\x00\xaa\xbb\xcc", b"test-string"])
@pytest.mark.parametrize("mode", ["644", "600", "755"])
@pytest.mark.parametrize(("user", "group"), [("root", "root"), ("ubuntu", "ubuntu")])
def test_push_file_io(reusable_instance, content, mode, user, group):
    reusable_instance.push_file_io(
        destination=pathlib.PurePosixPath("/tmp/create-file-test.txt"),
        content=io.BytesIO(content),
        file_mode=mode,
        user=user,
        group=group,
    )

    proc = reusable_instance.execute_run(
        command=["cat", "/tmp/create-file-test.txt"],
        capture_output=True,
    )

    assert proc.stdout == content

    proc = reusable_instance.execute_run(
        command=["stat", "--format", "%a:%U:%G", "/tmp/create-file-test.txt"],
        capture_output=True,
        text=True,
    )

    assert proc.stdout.strip() == f"{mode}:{user}:{group}"


@pytest.mark.parametrize(
    "destination",
    [
        pathlib.PurePosixPath("/tmp/push-file.txt"),
        pathlib.PurePosixPath("/home/ubuntu/push-file.txt"),
        pathlib.PurePosixPath("/root/push-file.txt"),
        pathlib.PurePosixPath("/push-file.txt"),
    ],
)
@pytest.mark.parametrize("mode", ["644", "600", "755"])
def test_push_file(destination, mode, reusable_instance, simple_file):
    """Push a file into a Multipass instance."""
    simple_file.chmod(int(mode, 8))

    reusable_instance.push_file(source=simple_file, destination=destination)

    # check file contents
    proc = reusable_instance.execute_run(
        command=["cat", str(destination)], capture_output=True
    )
    assert proc.stdout.decode() == "this is a test"
    # check file ownership
    proc = reusable_instance.execute_run(
        command=["stat", "--format", "%a:%U:%G", str(destination)],
        capture_output=True,
        text=True,
    )
    assert proc.stdout.strip() == f"{mode}:ubuntu:ubuntu"

    # clean up before the next test
    reusable_instance.execute_run(["rm", str(destination)])


@pytest.mark.parametrize(
    "destination",
    [
        pathlib.PurePosixPath("/"),
        pathlib.PurePosixPath("/home/ubuntu/"),
        pathlib.PurePosixPath("/root/"),
        pathlib.PurePosixPath("/tmp/"),
    ],
)
@pytest.mark.parametrize("mode", ["644", "600", "755"])
def test_push_file_to_directory(destination, mode, reusable_instance, simple_file):
    """Push a file to a directory in a multipass instance."""
    simple_file.chmod(int(mode, 8))
    final_destination = str(destination / simple_file.name)

    reusable_instance.push_file(source=simple_file, destination=destination)

    # check file contents
    proc = reusable_instance.execute_run(
        command=["cat", final_destination], capture_output=True
    )
    assert proc.stdout.decode() == "this is a test"
    # check file permissions
    proc = reusable_instance.execute_run(
        command=["stat", "--format", "%a:%U:%G", final_destination],
        capture_output=True,
        text=True,
    )
    assert proc.stdout.strip() == f"{mode}:ubuntu:ubuntu"

    # clean up before the next test
    reusable_instance.execute_run(["rm", final_destination])


def test_push_file_source_directory_error(reusable_instance, home_tmp_path):
    """Raise an error if the source is a directory."""
    with pytest.raises(IsADirectoryError) as exc_info:
        reusable_instance.push_file(
            source=home_tmp_path,
            destination=pathlib.PurePosixPath("/path/to/non/existent/directory"),
        )

    assert str(exc_info.value) == (
        f"Source cannot be a directory: {str(home_tmp_path)!r}"
    )


def test_push_file_no_source_error(reusable_instance, home_tmp_path):
    """Raise an error if the source file does not exist."""
    source = home_tmp_path / "src.txt"

    with pytest.raises(FileNotFoundError) as exc_info:
        reusable_instance.push_file(
            source=source,
            destination=pathlib.PurePosixPath("/path/to/non/existent/directory"),
        )

    assert str(exc_info.value) == f"File not found: {str(source)!r}"


def test_push_file_no_parent_directory_error(reusable_instance, simple_file):
    """Raise an error if the parent directory of the destination does not exist."""
    with pytest.raises(FileNotFoundError) as exc_info:
        reusable_instance.push_file(
            source=simple_file,
            destination=pathlib.PurePosixPath("/path/to/non/existent/directory"),
        )

    assert str(exc_info.value) == (
        "Directory not found in instance: '/path/to/non/existent'"
    )


def test_delete(instance):
    assert instance.exists() is True

    instance.delete()

    assert instance.exists() is False


def test_execute_popen(reusable_instance):
    with reusable_instance.execute_popen(
        command=["pwd"],
        stdout=subprocess.PIPE,
        text=True,
    ) as proc:
        stdout, _ = proc.communicate()

    assert stdout.strip() == "/home/ubuntu"


def test_execute_popen_cwd(reusable_instance):
    with reusable_instance.execute_popen(
        command=["pwd"],
        cwd=pathlib.PurePosixPath("/"),
        stdout=subprocess.PIPE,
        text=True,
    ) as proc:
        stdout, _ = proc.communicate()

    assert stdout.strip() == "/"


def test_execute_run(reusable_instance):
    proc = reusable_instance.execute_run(
        command=["pwd"],
        capture_output=True,
        text=True,
    )

    assert proc.stdout.strip() == "/home/ubuntu"


def test_execute_run_cwd(reusable_instance):
    proc = reusable_instance.execute_run(
        command=["pwd"],
        cwd=pathlib.PurePosixPath("/"),
        capture_output=True,
        text=True,
    )

    assert proc.stdout.strip() == "/"


def test_exists(reusable_instance):
    assert reusable_instance.exists() is True


def test_exists_false():
    fake_instance = MultipassInstance(name="does-not-exist")

    assert fake_instance.exists() is False


def test_launch(instance_name):
    instance = MultipassInstance(name=instance_name)

    assert instance.exists() is False

    instance.launch(
        image="snapcraft:core22",
        cpus=4,
        disk_gb=128,
        mem_gb=1,
    )

    assert instance.exists() is True


def test_mount_unmount(reusable_instance, home_tmp_path):
    host_source = home_tmp_path
    target = pathlib.PurePosixPath("/tmp/mnt")

    test_file = host_source / "test.txt"
    test_file.write_text("this is a test")

    assert reusable_instance.is_mounted(host_source=host_source, target=target) is False

    reusable_instance.mount(host_source=host_source, target=target)

    assert reusable_instance.is_mounted(host_source=host_source, target=target) is True

    proc = reusable_instance.execute_run(
        command=["cat", "/tmp/mnt/test.txt"],
        capture_output=True,
    )

    assert proc.stdout == test_file.read_bytes()

    reusable_instance.unmount(target=target)

    assert reusable_instance.is_mounted(host_source=host_source, target=target) is False


def test_mount_unmount_all(reusable_instance, home_tmp_path):
    host_source_1 = home_tmp_path / "1"
    host_source_1.mkdir()
    target_1 = pathlib.PurePosixPath("/tmp/mnt/1")

    host_source_2 = home_tmp_path / "2"
    host_source_2.mkdir()
    target_2 = pathlib.PurePosixPath("/tmp/mnt/2")

    reusable_instance.mount(host_source=host_source_1, target=target_1)
    reusable_instance.mount(host_source=host_source_2, target=target_2)

    assert (
        reusable_instance.is_mounted(host_source=host_source_1, target=target_1) is True
    )
    assert (
        reusable_instance.is_mounted(host_source=host_source_2, target=target_2) is True
    )

    reusable_instance.unmount_all()

    assert (
        reusable_instance.is_mounted(host_source=host_source_1, target=target_1)
        is False
    )
    assert (
        reusable_instance.is_mounted(host_source=host_source_2, target=target_2)
        is False
    )


def test_start_stop_is_running(reusable_instance):
    assert reusable_instance.is_running() is True

    reusable_instance.stop()

    assert reusable_instance.is_running() is False

    reusable_instance.start()

    assert reusable_instance.is_running() is True



#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import contextlib
import shutil
from pathlib import Path

import pytest


@pytest.fixture()
def fake_executor_local_pull(fake_executor):
    """Provide an executor that copies the file locally on 'pull'."""

    def copy_file(source, destination):
        shutil.copy(source, destination)

    fake_executor.pull_file = copy_file
    return fake_executor


@pytest.fixture()
def mock_home_temp_file(mocker, tmp_path):
    """Mock `home_temporary_file()`."""

    @contextlib.contextmanager
    def _mock_home_temp_file():
        tmp_file = Path(tmp_path / "fake-temp-file.txt")
        try:
            yield tmp_file
        finally:
            tmp_file.unlink()

    return mocker.patch(
        "craft_providers.util.temp_paths.home_temporary_file",
        wraps=_mock_home_temp_file,
    )


def test_temporarypull_ok(
    mock_home_temp_file, mocker, tmp_path, fake_executor_local_pull
):
    """Successful case."""
    source = tmp_path / "source.txt"
    source.write_text("test content")

    with fake_executor_local_pull.temporarily_pull_file(source=source) as localfilepath:
        # content copied and accessible
        assert localfilepath.read_text() == "test content"

    # file is removed afterwards
    assert not localfilepath.exists()

    # temp_paths.home_temporary_file() is used for the local file
    mock_home_temp_file.assert_called_once()


def test_temporarypull_missing_file_error(tmp_path, fake_executor_local_pull):
    """The source is missing, by default it's an error."""
    source = tmp_path / "source.txt"  # note we're not creating it in disk

    with pytest.raises(FileNotFoundError):
        with fake_executor_local_pull.temporarily_pull_file(source=source):
            pass


def test_temporarypull_missing_file_ok(tmp_path, fake_executor_local_pull):
    """The source is missing, but it's ok."""
    source = tmp_path / "source.txt"  # note we're not creating it in disk

    with fake_executor_local_pull.temporarily_pull_file(
        source=source, missing_ok=True
    ) as localfilepath:
        assert localfilepath is None


def test_temporarypull_temp_file_cleaned(
    mock_home_temp_file, tmp_path, fake_executor_local_pull
):
    """The temp file is cleaned after usage."""
    source = tmp_path / "source.txt"
    source.write_text("test content")

    with pytest.raises(ValueError), fake_executor_local_pull.temporarily_pull_file(
        source=source
    ) as localfilepath:
        # internal crash
        raise ValueError("boom")

    # file is removed afterwards
    assert not localfilepath.exists()  # pyright: ignore [reportUnboundVariable]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import pytest
from craft_providers.bases import almalinux, get_base_from_alias, ubuntu
from craft_providers.lxd import LXDProvider, is_installed


def test_ensure_provider_is_available_not_installed(uninstalled_lxd):
    """Verify lxd is installed and configured."""
    assert is_installed() is False
    provider = LXDProvider()
    provider.ensure_provider_is_available()


def test_ensure_provider_is_available_installed(installed_lxd):
    """Verify lxd is installed and configured."""
    assert is_installed() is True
    provider = LXDProvider()
    provider.ensure_provider_is_available()


def test_create_environment(installed_lxd, instance_name):
    provider = LXDProvider()
    test_instance = provider.create_environment(instance_name=instance_name)
    assert test_instance.exists() is False


@pytest.mark.parametrize(
    "alias",
    set(ubuntu.BuilddBaseAlias) - {ubuntu.BuilddBaseAlias.XENIAL}
    | {almalinux.AlmaLinuxBaseAlias.NINE},
)
def test_launched_environment(
    alias, installed_lxd, instance_name, tmp_path, session_provider
):
    cache_path = tmp_path / "cache"
    project_path = tmp_path / "project"
    cache_path.mkdir()
    project_path.mkdir()

    base_configuration = get_base_from_alias(alias)(alias=alias, cache_path=cache_path)

    with session_provider.launched_environment(
        project_name="test-project",
        project_path=project_path,
        base_configuration=base_configuration,
        instance_name=instance_name,
        allow_unstable=True,
    ) as test_instance:
        assert test_instance.exists() is True
        assert test_instance.is_running() is True
        test_instance.execute_run(["touch", "/root/.cache/pip/test-pip-cache"])

        assert (
            cache_path
            / base_configuration.compatibility_tag
            / str(base_configuration.alias)
            / "pip"
            / "test-pip-cache"
        ).exists()

    assert test_instance.exists() is True
    assert test_instance.is_running() is False



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""LXD Errors."""

from typing import Optional

from craft_providers.errors import ProviderError

LXD_INSTALL_HELP = (
    "Visit https://documentation.ubuntu.com/lxd/en/latest/getting_started/"
    " for instructions on installing and configuring LXD for your operating system."
)


class LXDError(ProviderError):
    """Unexpected LXD error."""


class LXDInstallationError(LXDError):
    """LXD Installation Error.

    :param reason: Reason for install failure.
    :param details: Optional details to include.
    """

    def __init__(
        self,
        reason: str,
        *,
        details: Optional[str] = None,
    ) -> None:
        brief = f"Failed to install LXD: {reason}."
        resolution = LXD_INSTALL_HELP
        super().__init__(brief=brief, details=details, resolution=resolution)


class LXDUnstableImageError(LXDError):
    """LXD Unstable Image Error.

    :param brief: Brief description of error.
    """

    def __init__(self, brief: str) -> None:
        super().__init__(
            brief=brief,
            details=(
                "Devel or daily images are not guaranteed and are intended for "
                "experimental use only."
            ),
        )



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import shutil

import pytest
from craft_providers import lxd


def test_is_initialized(installed_lxd):
    assert lxd.is_initialized(lxc=lxd.LXC(), remote="local") is True


def test_is_initialized_false(installed_lxd_without_init):
    assert lxd.is_initialized(lxc=lxd.LXC(), remote="local") is False


def test_is_installed():
    expected = shutil.which("lxd") is not None

    assert lxd.is_installed() is expected


def test_install(uninstalled_lxd):
    assert lxd.is_installed() is False

    lxd_version = lxd.install()

    assert lxd.is_installed() is True
    assert lxd_version is not None


def test_ensure_lxd_is_ready(installed_lxd_without_init):
    with pytest.raises(lxd.LXDError) as exc_info:
        lxd.ensure_lxd_is_ready()

    assert exc_info.value == lxd.LXDError(
        brief="LXD has not been properly initialized.",
        resolution="Execute 'lxd init --auto' to initialize LXD.\n"
        "Visit https://documentation.ubuntu.com/lxd/en/latest/getting_started/ for "
        "instructions on installing and configuring LXD for your operating system.",
    )



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Helpers for snap commands."""

import contextlib
import json
import logging
import pathlib
import shlex
import subprocess
import urllib.parse
from typing import Any, Dict, Iterator, List, Optional

import pydantic
import requests
import requests_unixsocket  # type: ignore

from craft_providers.const import TIMEOUT_COMPLEX, TIMEOUT_SIMPLE
from craft_providers.errors import (
    BaseConfigurationError,
    ProviderError,
    details_from_called_process_error,
)
from craft_providers.executor import Executor
from craft_providers.instance_config import InstanceConfiguration
from craft_providers.util import snap_cmd, temp_paths

logger = logging.getLogger(__name__)


# possible sources for the snap (using these two constants instead
# of an enum because the values are persisted with JSON)
SNAP_SRC_HOST = "host"
SNAP_SRC_STORE = "store"


class SnapInstallationError(ProviderError):
    """Unexpected error during snap installation."""


class Snap(pydantic.BaseModel, extra=pydantic.Extra.forbid):
    """Details of snap to install in the base.

    :param name: name of snap
    :param channel: snap store channel to install from (default is stable)
      If channel is `None`, then the snap is injected from the host instead
      of being installed from the store.
    :param classic: true if snap is a classic snap (default is false)
    """

    name: str
    channel: Optional[str] = "stable"
    classic: bool = False

    @pydantic.validator("channel")
    def validate_channel(cls, channel):
        """Validate that channel is not an empty string.

        :raises BaseConfigurationError: if channel is empty
        """
        if channel is not None and not channel:
            raise BaseConfigurationError(
                brief="channel cannot be empty",
                resolution="set channel to a non-empty string or `None`",
            )
        return channel


def _download_host_snap(
    *, snap_name: str, output: pathlib.Path, chunk_size: int = 64 * 1024
) -> None:
    """Download the current host snap using snapd's APIs."""
    quoted_name = urllib.parse.quote(snap_name, safe="")
    url = f"http+unix://%2Frun%2Fsnapd.socket/v2/snaps/{quoted_name}/file"
    try:
        resp = requests_unixsocket.get(url)
    except requests.exceptions.ConnectionError as error:
        raise SnapInstallationError(
            brief="Unable to connect to snapd service."
        ) from error

    try:
        resp.raise_for_status()
    except requests.exceptions.HTTPError as error:
        raise SnapInstallationError(
            brief=f"Unable to download snap {snap_name!r} from snapd."
        ) from error

    with output.open("wb") as stream:
        for chunk in resp.iter_content(chunk_size):
            stream.write(chunk)


def _pack_host_snap(*, snap_name: str, output: pathlib.Path) -> None:
    """Pack the current host snap."""
    command = snap_cmd.formulate_pack_command(snap_name, output)

    logger.debug("Executing command on host: %s", shlex.join(command))
    subprocess.run(
        command,
        capture_output=True,
        check=True,
    )


def get_host_snap_info(snap_name: str) -> Dict[str, Any]:
    """Get info about a snap installed on the host."""
    quoted_name = urllib.parse.quote(snap_name, safe="")
    url = f"http+unix://%2Frun%2Fsnapd.socket/v2/snaps/{quoted_name}"
    try:
        snap_info = requests_unixsocket.get(url)
    except requests.exceptions.ConnectionError as error:
        raise SnapInstallationError(
            brief="Unable to connect to snapd service."
        ) from error
    snap_info.raise_for_status()
    # TODO: represent snap info in a dataclass
    return snap_info.json()["result"]


def _get_target_snap_revision_from_snapd(
    snap_name: str, executor: Executor
) -> Optional[str]:
    """Get the revision of the snap on the target."""
    quoted_name = urllib.parse.quote(snap_name, safe="")
    url = f"http://localhost/v2/snaps/{quoted_name}"
    cmd = ["curl", "--silent", "--unix-socket", "/run/snapd.socket", url]
    try:
        proc = executor.execute_run(
            cmd, check=True, capture_output=True, timeout=TIMEOUT_SIMPLE
        )
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief="Unable to get target snap revision.",
            details=details_from_called_process_error(error),
        ) from error

    result = json.loads(proc.stdout)
    if result["status-code"] == 404:
        # snap not found
        return None
    if result["status-code"] == 200:
        return result["result"]["revision"]
    raise SnapInstallationError(f"Unknown response from snapd: {result!r}")


def _get_snap_revision_ensuring_source(
    snap_name: str, source: str, executor: Executor
) -> Optional[str]:
    """Get revision of snap on target and ensure the installation source."""
    instance_config = InstanceConfiguration.load(executor=executor)
    if instance_config is None or instance_config.snaps is None:
        return None

    config = instance_config.snaps.get(snap_name)
    if config is None:
        # not installed before
        return None

    # use 'get' to retrieve the source to support configs
    # saved by previous versions of the lib
    if config.get("source") == source:
        # previously installed from specified source: ok
        return config["revision"]

    # installed from other source: remove it
    logger.debug(
        "Snap %r installed from other source (%s), removing", snap_name, config
    )
    cmd = snap_cmd.formulate_remove_command(snap_name)
    try:
        executor.execute_run(
            cmd, check=True, capture_output=True, timeout=TIMEOUT_SIMPLE
        )
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief=f"Failed to remove snap {snap_name!r}.",
            details=details_from_called_process_error(error),
        ) from error
    return None


@contextlib.contextmanager
def _get_host_snap(snap_name: str) -> Iterator[pathlib.Path]:
    """Get snap installed on host containing the config.

    Snapd provides an API to fetch a snap. First use that to fetch a snap.
    If the snap is installed using `snap try`, it may fail to download. In
    that case, attempt to construct the snap by packing it ourselves.

    :yields: context manager that sets the temporary snap installation file
      as the target
    """
    with temp_paths.home_temporary_directory() as tmp_dir:
        snap_path = tmp_dir / f"{snap_name}.snap"
        try:
            _download_host_snap(snap_name=snap_name, output=snap_path)
        except SnapInstallationError:
            logger.debug(
                "Failed to fetch snap from snapd,"
                " falling back to `snap pack` to recreate"
            )
            _pack_host_snap(snap_name=snap_name, output=snap_path)

        yield snap_path


def _get_assertion(query: List[str]) -> bytes:
    """Get an assertion from snapd.

    :param query: assertion query to pass to `snap known`
    :returns: assertion data
    :raises SnapInstallationError: if 'snap known' call fails
    """
    command = snap_cmd.formulate_known_command(query=query)
    logger.debug("Executing command on host: %s", command)
    try:
        return subprocess.run(command, capture_output=True, check=True).stdout
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief="failed to get assertions for snap",
            details=details_from_called_process_error(error),
        ) from error


@contextlib.contextmanager
def _get_assertions_file(
    snap_name: str, snap_id: str, snap_revision: str, snap_publisher_id: str
) -> Iterator[pathlib.Path]:
    """Get an assertion file for a snap.

    :param snap_name: Name of snap to inject
    :param snap_id: ID of the snap
    :param snap_revision: Revision of the snap
    :param snap_publisher_id: The ID of the snap's publisher's account

    :yields: context manager that will set the temporary snap assertion file
      as the target
    """
    logger.debug("Creating an assert file for snap %r", snap_name)
    assertion_queries = [
        [
            "account-key",
            "public-key-sha3-384=BWDEoaqyr25nF5SNCvEv2v"
            "7QnM9QsfCc0PBMYD_i2NGSQ32EF2d4D0hqUel3m8ul",
        ],
        ["snap-declaration", f"snap-name={snap_name}"],
        ["snap-revision", f"snap-revision={snap_revision}", f"snap-id={snap_id}"],
        ["account", f"account-id={snap_publisher_id}"],
    ]

    with temp_paths.home_temporary_file() as assert_file_path:
        with open(assert_file_path, "wb") as assert_file:
            for query in assertion_queries:
                assert_file.write(_get_assertion(query))
                assert_file.write(b"\n")
            assert_file.flush()
            yield assert_file_path


def _add_assertions_from_host(executor: Executor, snap_name: str) -> None:
    """Add assertions from the host into the target for a snap.

    :param executor: Executor for target
    :param snap_name: Name of snap to inject
    """
    # trim the `_name` suffix, if present
    target_assert_path = pathlib.PurePosixPath(f"/tmp/{snap_name.split('_')[0]}.assert")
    snap_info = get_host_snap_info(snap_name)

    try:
        with _get_assertions_file(
            snap_name=snap_name,
            snap_id=snap_info["id"],
            snap_revision=snap_info["revision"],
            snap_publisher_id=snap_info["publisher"]["id"],
        ) as host_assert_path:
            executor.push_file(
                source=host_assert_path,
                destination=target_assert_path,
            )
    except ProviderError as error:
        raise SnapInstallationError(
            brief=f"failed to copy assert file for snap {snap_name!r}",
            details="error copying snap assert file into target environment",
        ) from error

    try:
        executor.execute_run(
            snap_cmd.formulate_ack_command(snap_assert_path=target_assert_path),
            check=True,
            capture_output=True,
            timeout=TIMEOUT_COMPLEX,
        )
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief=f"failed to add assertions for snap {snap_name!r}",
            details=details_from_called_process_error(error),
        ) from error


def inject_from_host(*, executor: Executor, snap_name: str, classic: bool) -> None:
    """Inject snap from host snap.

    If the snap on the host was installed with the `--name` parameter, the host snap's
    name will be formatted as `<snap_name>_<name>`. The `--name` parameter is not
    used when the snap is installed inside the instance, so the instance's snap
    name will just be `<snap_name>` (no suffix).

    :param executor: Executor for target
    :param snap_name: Name of snap to inject
    :param classic: Install in classic mode

    :raises SnapInstallationError: on failure to inject snap
    """
    # the local snap name may have a suffix if it was installed with `--name`
    snap_store_name = snap_name.split("_")[0]
    if snap_name == snap_store_name:
        logger.debug("Installing snap %r from host (classic=%s)", snap_name, classic)
    else:
        logger.debug(
            "Installing snap %r from host as %r in instance (classic=%s).",
            snap_name,
            snap_store_name,
            classic,
        )

    host_revision = get_host_snap_info(snap_name)["revision"]
    target_revision = _get_snap_revision_ensuring_source(
        snap_name=snap_store_name,
        source=SNAP_SRC_HOST,
        executor=executor,
    )
    logger.debug("Revisions found: host=%r, target=%r", host_revision, target_revision)

    if target_revision is not None and target_revision == host_revision:
        logger.debug(
            "Skipping snap injection:"
            " target is already up-to-date with revision on host"
        )
        return

    target_snap_path = pathlib.PurePosixPath(f"/tmp/{snap_store_name}.snap")
    is_dangerous = host_revision.startswith("x")

    if not is_dangerous:
        _add_assertions_from_host(executor=executor, snap_name=snap_name)

    with _get_host_snap(snap_name) as host_snap_path:
        try:
            executor.push_file(
                source=host_snap_path,
                destination=target_snap_path,
            )
        except ProviderError as error:
            raise SnapInstallationError(
                brief=f"failed to copy snap file for snap {snap_name!r}",
                details="error copying snap file into target environment",
            ) from error

    try:
        executor.execute_run(
            snap_cmd.formulate_local_install_command(
                classic=classic, dangerous=is_dangerous, snap_path=target_snap_path
            ),
            check=True,
            capture_output=True,
            timeout=TIMEOUT_COMPLEX,
        )
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief=f"failed to install snap {snap_store_name!r}",
            details=details_from_called_process_error(error),
        ) from error

    InstanceConfiguration.update(
        executor=executor,
        data={
            "snaps": {
                snap_store_name: {"revision": host_revision, "source": SNAP_SRC_HOST}
            }
        },
    )


def install_from_store(
    *, executor: Executor, snap_name: str, channel: str, classic: bool
) -> None:
    """Install snap from store into target.

    Perform installation using method which prevents refreshing.

    :param executor: Executor for target.
    :param snap_name: Name of snap to install.
    :param channel: Channel to install from.
    :param classic: Install in classic mode.

    :raises SnapInstallationError: on unexpected error.
    """
    # trim the `_name` suffix, if present
    snap_store_name = snap_name.split("_")[0]
    if snap_name == snap_store_name:
        logger.debug(
            "Installing snap %r from store (channel=%r, classic=%s).",
            snap_name,
            channel,
            classic,
        )
    else:
        logger.debug(
            "Installing snap %r as %r from store (channel=%r, classic=%s).",
            snap_name,
            snap_store_name,
            channel,
            classic,
        )

    target_revision = _get_snap_revision_ensuring_source(
        snap_name=snap_store_name,
        source=SNAP_SRC_STORE,
        executor=executor,
    )
    logger.debug("Revision found in target: %r", target_revision)

    if target_revision is None:
        # no snap present in the target environment, just install it
        cmd = snap_cmd.formulate_remote_install_command(
            snap_name=snap_store_name,
            channel=channel,
            classic=classic,
        )
    else:
        # refresh the already installed snap
        cmd = snap_cmd.formulate_refresh_command(
            snap_name=snap_store_name,
            channel=channel,
        )

    try:
        executor.execute_run(
            cmd,
            check=True,
            capture_output=True,
            timeout=TIMEOUT_COMPLEX,
        )
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief=f"Failed to install/refresh snap {snap_store_name!r}.",
            details=details_from_called_process_error(error),
        ) from error

    new_target_revision = _get_target_snap_revision_from_snapd(
        snap_name=snap_store_name,
        executor=executor,
    )
    logger.debug("Revision after install/refresh: %r", new_target_revision)

    InstanceConfiguration.update(
        executor=executor,
        data={
            "snaps": {
                snap_store_name: {
                    "revision": new_target_revision,
                    "source": SNAP_SRC_STORE,
                }
            }
        },
    )



#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Tests for abstract Base's implementations."""
import enum
import pathlib
import subprocess
import sys
from unittest import mock

import pytest
import pytest_subprocess.fake_popen
from craft_providers import Executor, base
from craft_providers.errors import BaseConfigurationError

from tests.unit.conftest import DEFAULT_FAKE_CMD

pytestmark = [pytest.mark.usefixtures("instant_sleep")]

WAIT_FOR_SYSTEM_READY_CMD = ["systemctl", "is-system-running"]
WAIT_FOR_NETWORK_CMD = ["getent", "hosts", "snapcraft.io"]


class FakeBase(base.Base):
    FakeBaseAlias = enum.Enum("FakeBaseAlias", ["TREBLE"])
    _environment = {}

    _hostname = "my-hostname"
    _retry_wait = 0.01

    # Very small retry and timeout values so tests don't take long.
    _timeout_simple = 1
    _timeout_complex = _timeout_simple * 2
    _timeout_unpredictable = _timeout_complex * 2
    alias = FakeBaseAlias.TREBLE

    # Minimal implementations of abstract methods
    def __init__(self, **kwargs):
        pass

    def _ensure_os_compatible(self, executor: Executor) -> None:
        pass

    def _setup_packages(self, executor: Executor) -> None:
        pass

    def _setup_snapd(self, executor: Executor) -> None:
        pass


@pytest.fixture()
def fake_base() -> base.Base:
    return FakeBase()


@pytest.fixture()
def mock_executor():
    return mock.Mock(spec=Executor)


@pytest.fixture(params=[None, 0.1, 1])
def timeout_value(request):
    return request.param


def raise_timeout(process: pytest_subprocess.fake_popen.FakePopen):
    """Raise a TimeoutExpired exception, for use as a callback for subprocess."""
    raise subprocess.TimeoutExpired(process.args, 0)


@pytest.mark.parametrize("running_state", ["running", "degraded"])
def test_wait_for_system_ready_success(
    fake_base, fake_executor, fake_process, running_state, failure_count, timeout_value
):
    fake_base._timeout_simple = timeout_value
    cmd = [*DEFAULT_FAKE_CMD, *WAIT_FOR_SYSTEM_READY_CMD]
    for _ in range(failure_count):
        fake_process.register(cmd, stdout="no")
    fake_process.register(cmd, stdout=running_state)

    fake_base._setup_wait_for_system_ready(fake_executor)


@pytest.mark.parametrize(
    "callback",
    [
        pytest.param(raise_timeout, id="subprocess-timeout"),
        pytest.param(lambda _: None, id="loop-timeout"),
    ],
)
def test_wait_for_system_ready_timeout(
    fake_base, fake_executor, fake_process, callback
):
    fake_process.register(
        [*DEFAULT_FAKE_CMD, *WAIT_FOR_SYSTEM_READY_CMD], callback=callback
    )
    fake_process.keep_last_process(True)

    with pytest.raises(BaseConfigurationError):
        fake_base._setup_wait_for_system_ready(fake_executor)


def test_wait_for_network_success(
    fake_base, fake_executor, fake_process, failure_count, timeout_value
):
    fake_base._timeout_simple = timeout_value
    cmd = [*DEFAULT_FAKE_CMD, *WAIT_FOR_NETWORK_CMD]
    for _ in range(failure_count):
        fake_process.register(cmd, returncode=1)
    fake_process.register(cmd, returncode=0)

    fake_base._setup_wait_for_network(fake_executor)


@pytest.mark.parametrize(
    "callback",
    [
        pytest.param(raise_timeout, id="subprocess-timeout"),
        pytest.param(lambda _: None, id="loop-timeout"),
    ],
)
def test_wait_for_network_timeout(fake_base, fake_executor, fake_process, callback):
    fake_process.register([*DEFAULT_FAKE_CMD, *WAIT_FOR_NETWORK_CMD], callback=callback)
    fake_process.keep_last_process(True)

    with pytest.raises(BaseConfigurationError):
        fake_base._setup_wait_for_system_ready(fake_executor)


@pytest.mark.parametrize("cache_dir", [pathlib.Path("/tmp/fake-cache-dir")])
def test_mount_shared_cache_dirs(fake_process, fake_base, fake_executor, cache_dir):
    """Test mounting of cache directories with a cache directory set."""
    fake_base._cache_path = cache_dir
    user_cache_dir = pathlib.Path("/root/.cache")

    fake_process.register(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "echo -n ${XDG_CACHE_HOME:-${HOME}/.cache}"],
        stdout=str(user_cache_dir),
    )

    fake_process.register(
        [*DEFAULT_FAKE_CMD, "mkdir", "-p", "/root/.cache/pip"],
    )

    fake_base._mount_shared_cache_dirs(fake_executor)

    if sys.platform == "win32":
        expected = {
            "host_source": pathlib.WindowsPath("d:")
            / cache_dir
            / "base-v2"
            / "FakeBaseAlias.TREBLE"
            / "pip",
            "target": user_cache_dir / "pip",
        }
    else:
        expected = {
            "host_source": cache_dir / "base-v2" / "FakeBaseAlias.TREBLE" / "pip",
            "target": user_cache_dir / "pip",
        }
    assert fake_executor.records_of_mount == [expected]


@pytest.mark.parametrize("cache_dir", [pathlib.Path("/tmp/fake-cache-dir")])
def test_mount_shared_cache_dirs_mkdir_failed(
    fake_process, fake_base, fake_executor, cache_dir, mocker
):
    """Test mounting of cache directories with a cache directory set, but mkdir failed."""
    fake_base._cache_path = cache_dir
    user_cache_dir = pathlib.Path("/root/.cache")

    fake_process.register(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "echo -n ${XDG_CACHE_HOME:-${HOME}/.cache}"],
        stdout=str(user_cache_dir),
    )

    fake_process.register(
        [*DEFAULT_FAKE_CMD, "mkdir", "-p", "/root/.cache/pip"],
    )

    mocker.patch("pathlib.Path.mkdir", side_effect=OSError)

    with pytest.raises(BaseConfigurationError):
        fake_base._mount_shared_cache_dirs(fake_executor)


@pytest.mark.parametrize(
    ("process_outputs", "expected"),
    [
        (
            [
                'NAME="AlmaLinux"\nVERSION="9.1 (Lime Lynx)"\nID="almalinux"\nID_LIKE="rhel centos fedora"\nVERSION_ID="9.1"\n'
            ],
            {
                "NAME": "AlmaLinux",
                "ID": "almalinux",
                "VERSION": "9.1 (Lime Lynx)",
                "VERSION_ID": "9.1",
                "ID_LIKE": "rhel centos fedora",
            },
        ),
        (
            [
                "",
                'NAME="AlmaLinux"\nVERSION="9.1 (Lime Lynx)"\nID="almalinux"\nID_LIKE="rhel centos fedora"\nVERSION_ID="9.1"\n',
            ],
            {
                "NAME": "AlmaLinux",
                "ID": "almalinux",
                "VERSION": "9.1 (Lime Lynx)",
                "VERSION_ID": "9.1",
                "ID_LIKE": "rhel centos fedora",
            },
        ),
        (
            [
                'NAME="CentOS Linux"\nVERSION="7 (Core)"\nID="centos"\nID_LIKE="rhel fedora"\nVERSION_ID="7"\n'
            ],
            {
                "NAME": "CentOS Linux",
                "ID": "centos",
                "VERSION": "7 (Core)",
                "VERSION_ID": "7",
                "ID_LIKE": "rhel fedora",
            },
        ),
        (
            [
                "",
                'NAME="CentOS Linux"\nVERSION="7 (Core)"\nID="centos"\nID_LIKE="rhel fedora"\nVERSION_ID="7"\n',
            ],
            {
                "NAME": "CentOS Linux",
                "ID": "centos",
                "VERSION": "7 (Core)",
                "VERSION_ID": "7",
                "ID_LIKE": "rhel fedora",
            },
        ),
        (
            ["NAME=Ubuntu\nVERSION_ID=22.04\n"],
            {"NAME": "Ubuntu", "VERSION_ID": "22.04"},
        ),
    ],
)
def test_get_os_release_success(
    fake_process, fake_executor, fake_base, process_outputs, expected
):
    """`_get_os_release` should parse data from `/etc/os-release` to a dict."""
    for output in process_outputs:
        fake_process.register_subprocess(
            [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
            stdout=output,
        )

    result = fake_base._get_os_release(executor=fake_executor)

    assert result == expected


@pytest.mark.parametrize(
    ("stdout", "returncode"),
    [
        ("", 0),
        ("NAME=Linux", 1),
    ],
)
def test_get_os_release_error_output(
    fake_process, fake_executor, fake_base, stdout, returncode
):
    """Test when output values /etc/os-release are invalid."""
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=stdout,
        returncode=returncode,
    )
    fake_process.keep_last_process(True)

    with pytest.raises(BaseConfigurationError):
        fake_base._get_os_release(executor=fake_executor)


def test_get_os_release_exception(fake_base, mock_executor):
    mock_executor.execute_run.side_effect = subprocess.CalledProcessError(
        cmd=[], returncode=1
    )

    with pytest.raises(BaseConfigurationError):
        fake_base._get_os_release(executor=mock_executor)



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import copy
import io
import pathlib
import subprocess
import sys
from unittest import mock

import pytest
from craft_providers import errors
from craft_providers.multipass import Multipass, MultipassInstance
from craft_providers.multipass.errors import MultipassError

if sys.platform == "win32":
    EXAMPLE_MOUNTS = {
        "/root/project": {
            "gid_mappings": ["-2:default"],
            "source_path": "<insert>",
            "uid_mappings": ["-2:default"],
        }
    }
else:
    EXAMPLE_MOUNTS = {
        "/root/project": {
            "gid_mappings": ["1000:0"],
            "source_path": "<insert>",
            "uid_mappings": ["1000:0"],
        }
    }

EXAMPLE_INFO = {
    "errors": [],
    "info": {
        "flowing-hawfinch": {
            "disks": {"sda1": {}},
            "image_hash": (
                "c5f2f08c6a1adee1f2f96d84856bf0162d33ea182dae0e8ed45768a86182d110"
            ),
            "image_release": "22.04 LTS",
            "ipv4": [],
            "load": [],
            "memory": {},
            "mounts": {},
            "release": "",
            "state": "Stopped",
        },
        "test-instance": {
            "disks": {"sda1": {"total": "266219864064", "used": "1457451008"}},
            "image_hash": (
                "7c5c8f24046ca7b82897e0ca49fbd4dbdc771c2abd616991d10e6e09cc43002f"
            ),
            "image_release": "Snapcraft builder for Core 22",
            "ipv4": ["10.114.154.133"],
            "load": [1.53, 0.84, 0.33],
            "memory": {"total": 2089697280, "used": 153190400},
            "mounts": EXAMPLE_MOUNTS,
            "release": "Ubuntu 22.04.2 LTS",
            "state": "Running",
        },
    },
}


@pytest.fixture()
def project_path(tmp_path):
    project_path = tmp_path / "git" / "project"
    project_path.mkdir(parents=True)
    return project_path


@pytest.fixture(autouse=True)
def mock_multipass(project_path):
    with mock.patch(
        "craft_providers.multipass.multipass_instance.Multipass", spec=Multipass
    ) as multipass_mock:
        platform_info = copy.deepcopy(EXAMPLE_INFO)
        platform_info["info"]["test-instance"]["mounts"]["/root/project"][
            "source_path"
        ] = project_path.as_posix()

        multipass_mock.info.return_value = platform_info

        multipass_mock.list.return_value = ["flowing-hawfinch", "test-instance"]
        yield multipass_mock


@pytest.fixture()
def instance(mock_multipass):
    return MultipassInstance(name="test-instance", multipass=mock_multipass)


@pytest.fixture()
def simple_file(tmp_path):
    """Create a file in the test directory."""
    file = tmp_path / "src.txt"
    file.write_text("this is a test")
    return file


def test_push_file_io(mock_multipass, instance):
    mock_multipass.exec.side_effect = [
        mock.Mock(stdout="/tmp/mktemp-result\n"),
        None,
        None,
        None,
        None,
    ]

    instance.push_file_io(
        destination=pathlib.PurePosixPath("/etc/test.conf"),
        content=io.BytesIO(b"foo"),
        file_mode="0644",
    )

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "mktemp"],
            runner=subprocess.run,
            capture_output=True,
            check=True,
            text=True,
            timeout=60,
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=[
                "sudo",
                "-H",
                "--",
                "chown",
                "ubuntu:ubuntu",
                "/tmp/mktemp-result",
            ],
            runner=subprocess.run,
            capture_output=True,
            check=True,
            timeout=60,
        ),
        mock.call.transfer_source_io(
            source=mock.ANY, destination="test-instance:/tmp/mktemp-result"
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=[
                "sudo",
                "-H",
                "--",
                "chown",
                "root:root",
                "/tmp/mktemp-result",
            ],
            runner=subprocess.run,
            capture_output=True,
            check=True,
            timeout=60,
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "chmod", "0644", "/tmp/mktemp-result"],
            runner=subprocess.run,
            capture_output=True,
            check=True,
            timeout=60,
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=[
                "sudo",
                "-H",
                "--",
                "mv",
                "/tmp/mktemp-result",
                "/etc/test.conf",
            ],
            runner=subprocess.run,
            capture_output=True,
            check=True,
            timeout=600,
        ),
    ]


def test_push_file_io_error(mock_multipass, instance):
    error = subprocess.CalledProcessError(-1, ["mktemp"], "test stdout", "test stderr")

    mock_multipass.exec.side_effect = error

    with pytest.raises(MultipassError) as exc_info:
        instance.push_file_io(
            destination=pathlib.PurePosixPath("/etc/test.conf"),
            content=io.BytesIO(b"foo"),
            file_mode="0644",
        )

    assert exc_info.value == MultipassError(
        brief=(
            "Failed to create file '/etc/test.conf' in Multipass instance "
            "'test-instance'."
        ),
        details=errors.details_from_called_process_error(error),
    )


def test_delete(mock_multipass, instance):
    instance.delete()

    assert mock_multipass.mock_calls == [
        mock.call.delete(instance_name="test-instance", purge=True)
    ]


def test_execute_popen(mock_multipass, instance):
    instance.execute_popen(command=["test-command", "flags"], input="foo")

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test-command", "flags"],
            runner=subprocess.Popen,
            input="foo",
            timeout=None,
        )
    ]


def test_execute_popen_with_cwd(mock_multipass, instance):
    instance.execute_popen(
        command=["test-command", "flags"],
        cwd=pathlib.PurePosixPath("/tmp"),
        input="foo",
    )

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=[
                "sudo",
                "-H",
                "--",
                "env",
                "--chdir=/tmp",
                "test-command",
                "flags",
            ],
            runner=subprocess.Popen,
            input="foo",
            timeout=None,
        )
    ]


def test_execute_popen_with_env(mock_multipass, instance):
    instance.execute_popen(command=["test-command", "flags"], env={"foo": "bar"})

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "env", "foo=bar", "test-command", "flags"],
            runner=subprocess.Popen,
            timeout=None,
        )
    ]


def test_execute_run(mock_multipass, instance):
    instance.execute_run(command=["test-command", "flags"], input="foo")

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test-command", "flags"],
            runner=subprocess.run,
            input="foo",
            timeout=None,
            check=False,
        )
    ]


@pytest.mark.parametrize("check", [True, False])
def test_execute_run_with_check(check, mock_multipass, instance):
    instance.execute_run(command=["test-command", "flags"], input="foo", check=check)

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test-command", "flags"],
            runner=subprocess.run,
            input="foo",
            timeout=None,
            check=check,
        )
    ]


def test_execute_run_with_cwd(mock_multipass, instance, tmp_path):
    instance.execute_run(command=["test-command", "flags"], cwd=tmp_path)

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=[
                "sudo",
                "-H",
                "--",
                "env",
                f"--chdir={tmp_path.as_posix()}",
                "test-command",
                "flags",
            ],
            runner=subprocess.run,
            timeout=None,
            check=False,
        )
    ]


def test_execute_run_with_env(mock_multipass, instance):
    instance.execute_run(command=["test-command", "flags"], env={"foo": "bar"})

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "env", "foo=bar", "test-command", "flags"],
            runner=subprocess.run,
            timeout=None,
            check=False,
        )
    ]


def test_execute_run_with_env_unset(mock_multipass, instance):
    instance.execute_run(
        command=["test-command", "flags"], env={"foo": "bar", "TERM": None}
    )

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=[
                "sudo",
                "-H",
                "--",
                "env",
                "foo=bar",
                "-u",
                "TERM",
                "test-command",
                "flags",
            ],
            runner=subprocess.run,
            timeout=None,
            check=False,
        )
    ]


def test_exists(mock_multipass, instance):
    assert instance.exists() is True
    assert mock_multipass.mock_calls == [mock.call.list()]


def test_exists_false(mock_multipass):
    assert (
        MultipassInstance(name="does-not-exist", multipass=mock_multipass).exists()
        is False
    )
    assert mock_multipass.mock_calls == [mock.call.list()]


def test_is_mounted_false(mock_multipass, instance):
    project_path = pathlib.Path.home() / "not-mounted"

    assert (
        instance.is_mounted(
            host_source=pathlib.Path(project_path),
            target=pathlib.PurePosixPath("/root/project"),
        )
        is False
    )

    assert mock_multipass.mock_calls == [mock.call.info(instance_name="test-instance")]


def test_is_mounted_true(mock_multipass, instance, project_path):
    assert (
        instance.is_mounted(
            host_source=project_path,
            target=pathlib.PurePosixPath("/root/project"),
        )
        is True
    )

    assert mock_multipass.mock_calls == [mock.call.info(instance_name="test-instance")]


def test_is_running_false(mock_multipass):
    assert (
        MultipassInstance(
            name="flowing-hawfinch", multipass=mock_multipass
        ).is_running()
        is False
    )

    assert mock_multipass.mock_calls == [
        mock.call.info(instance_name="flowing-hawfinch")
    ]


def test_is_running_true(mock_multipass, instance):
    assert instance.is_running() is True

    assert mock_multipass.mock_calls == [mock.call.info(instance_name="test-instance")]


def test_launch(mock_multipass, instance):
    instance.launch(image="test-image")

    assert mock_multipass.mock_calls == [
        mock.call.launch(
            instance_name="test-instance",
            image="test-image",
            cpus="2",
            disk="256G",
            mem="2G",
        )
    ]


def test_launch_all_opts(mock_multipass, instance):
    instance.launch(image="test-image", cpus=4, disk_gb=5, mem_gb=6)

    assert mock_multipass.mock_calls == [
        mock.call.launch(
            instance_name="test-instance",
            image="test-image",
            cpus="4",
            disk="5G",
            mem="6G",
        )
    ]


def test_mount(mock_multipass, project_path):
    MultipassInstance(name="flowing-hawfinch", multipass=mock_multipass).mount(
        host_source=project_path,
        target=pathlib.PurePosixPath("/root/project"),
    )

    assert mock_multipass.mock_calls == [
        mock.call.info(instance_name="flowing-hawfinch"),
        mock.call.mount(
            source=project_path,
            target="flowing-hawfinch:/root/project",
        ),
    ]


def test_mount_already_mounted(mock_multipass, instance, project_path):
    instance.mount(
        host_source=project_path,
        target=pathlib.PurePosixPath("/root/project"),
    )

    assert mock_multipass.mock_calls == [mock.call.info(instance_name="test-instance")]


def test_pull_file(mock_multipass, instance, tmp_path):
    mock_multipass.exec.return_value = mock.Mock(returncode=0)

    source = pathlib.PurePosixPath("/tmp/src.txt")
    destination = tmp_path / "dst.txt"

    instance.pull_file(
        source=source,
        destination=destination,
    )

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test", "-f", "/tmp/src.txt"],
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
        mock.call.transfer(
            source="test-instance:/tmp/src.txt", destination=str(destination)
        ),
    ]


def test_pull_file_no_source(mock_multipass, instance, tmp_path):
    mock_multipass.exec.return_value = mock.Mock(returncode=1)

    source = pathlib.PurePosixPath("/tmp/src.txt")
    destination = tmp_path / "dst.txt"

    with pytest.raises(FileNotFoundError) as exc_info:
        instance.pull_file(
            source=source,
            destination=destination,
        )

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test", "-f", "/tmp/src.txt"],
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
    ]
    assert str(exc_info.value) == "File not found: '/tmp/src.txt'"


def test_pull_file_no_parent_directory(mock_multipass, instance, tmp_path):
    mock_multipass.exec.return_value = mock.Mock(returncode=0)

    source = pathlib.PurePosixPath("/tmp/src.txt")
    destination = tmp_path / "not-created" / "dst.txt"

    with pytest.raises(FileNotFoundError) as exc_info:
        instance.pull_file(
            source=source,
            destination=destination,
        )

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test", "-f", "/tmp/src.txt"],
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
    ]
    assert str(exc_info.value) == f"Directory not found: {str(destination.parent)!r}"


def test_push_file(mock_multipass, instance, simple_file):
    """Push a file into a Multipass instance."""
    mock_multipass.exec.side_effect = [
        # test call
        mock.Mock(returncode=0),
        # mktemp call
        mock.Mock(stdout="/tmp/mktemp-file\n", returncode=0),
        # chown call
        None,
        # test call (returncode=1 means the destination is not a directory)
        mock.Mock(returncode=1),
        # mv call
        None,
    ]

    destination = pathlib.PurePosixPath("/tmp/dst.txt")

    instance.push_file(source=simple_file, destination=destination)

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test", "-d", "/tmp"],
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "mktemp"],
            runner=subprocess.run,
            capture_output=True,
            check=True,
            text=True,
            timeout=60,
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "chown", "ubuntu:ubuntu", "/tmp/mktemp-file"],
            runner=subprocess.run,
            capture_output=True,
            check=True,
            timeout=60,
        ),
        mock.call.transfer(
            source=str(simple_file), destination="test-instance:/tmp/mktemp-file"
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test", "-d", "/tmp/dst.txt"],
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "mv", "/tmp/mktemp-file", "/tmp/dst.txt"],
            runner=subprocess.run,
            capture_output=True,
            check=True,
            timeout=None,
        ),
    ]


def test_push_file_to_directory(mock_multipass, instance, simple_file):
    """Push a file to a directory in a multipass instance."""
    mock_multipass.exec.side_effect = [
        # test call
        mock.Mock(returncode=0),
        # mktemp call
        mock.Mock(stdout="/tmp/mktemp-file\n", returncode=0),
        # chown call
        None,
        # test call (returncode=0 means the destination is a directory)
        mock.Mock(returncode=0),
        # mv call
        None,
    ]

    destination = pathlib.PurePosixPath("/tmp")

    instance.push_file(source=simple_file, destination=destination)

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test", "-d", "/"],
            runner=subprocess.run,
            timeout=60,
            check=False,
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "mktemp"],
            runner=subprocess.run,
            timeout=60,
            capture_output=True,
            check=True,
            text=True,
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "chown", "ubuntu:ubuntu", "/tmp/mktemp-file"],
            runner=subprocess.run,
            timeout=60,
            capture_output=True,
            check=True,
        ),
        mock.call.transfer(
            source=str(simple_file), destination="test-instance:/tmp/mktemp-file"
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test", "-d", "/tmp"],
            runner=subprocess.run,
            timeout=60,
            check=False,
        ),
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "mv", "/tmp/mktemp-file", "/tmp/src.txt"],
            runner=subprocess.run,
            timeout=None,
            capture_output=True,
            check=True,
        ),
    ]


def test_push_file_source_directory_error(mock_multipass, instance, tmp_path):
    """Raise an error if the source is a directory."""
    with pytest.raises(IsADirectoryError) as exc_info:
        instance.push_file(
            source=tmp_path, destination=pathlib.PurePosixPath("/tmp/dst.txt")
        )

    assert mock_multipass.mock_calls == []
    assert str(exc_info.value) == f"Source cannot be a directory: {str(tmp_path)!r}"


def test_push_file_no_source_error(mock_multipass, instance, tmp_path):
    """Raise an error if the source file does not exist."""
    source = tmp_path / "does-not-exist.txt"

    with pytest.raises(FileNotFoundError) as exc_info:
        instance.push_file(
            source=source, destination=pathlib.PurePosixPath("/tmp/dst.txt")
        )

    assert mock_multipass.mock_calls == []
    assert str(exc_info.value) == f"File not found: {str(source)!r}"


def test_push_file_no_parent_directory_error(mock_multipass, instance, simple_file):
    """Raise an error if the parent directory of the destination does not exist."""
    mock_multipass.exec.return_value = mock.Mock(returncode=1)

    with pytest.raises(FileNotFoundError) as exc_info:
        instance.push_file(
            source=simple_file,
            destination=pathlib.PurePosixPath("/tmp/dst.txt"),
        )

    assert mock_multipass.mock_calls == [
        mock.call.exec(
            instance_name="test-instance",
            command=["sudo", "-H", "--", "test", "-d", "/tmp"],
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
    ]
    assert str(exc_info.value) == "Directory not found in instance: '/tmp'"


def test_push_file_mktemp_error(mock_multipass, instance, simple_file):
    """Raise an error if the temporary file cannot be created."""
    error = subprocess.CalledProcessError(-1, ["mktemp"], "test stdout", "test stderr")

    mock_multipass.exec.side_effect = [
        # test call
        mock.Mock(returncode=0),
        # mktemp call
        error,
    ]

    with pytest.raises(MultipassError) as exc_info:
        instance.push_file(
            source=simple_file, destination=pathlib.PurePosixPath("/etc/test.conf")
        )

    assert exc_info.value == MultipassError(
        brief=(
            "Failed to push file '/etc/test.conf' into Multipass instance "
            "'test-instance'."
        ),
        details=errors.details_from_called_process_error(error),
    )


def test_push_file_chmod_error(mock_multipass, instance, simple_file):
    """Raise an error if the chmod call fails"""
    error = subprocess.CalledProcessError(-1, ["chmod"], "test stdout", "test stderr")

    mock_multipass.exec.side_effect = [
        # test call
        mock.Mock(returncode=0),
        # mktemp call
        mock.Mock(stdout="/tmp/mktemp-file\n", returncode=0),
        # chmod call
        error,
    ]

    with pytest.raises(MultipassError) as exc_info:
        instance.push_file(
            source=simple_file, destination=pathlib.PurePosixPath("/etc/test.conf")
        )

    assert exc_info.value == MultipassError(
        brief=(
            "Failed to push file '/etc/test.conf' into Multipass instance "
            "'test-instance'."
        ),
        details=errors.details_from_called_process_error(error),
    )


def test_push_file_mv_error(mock_multipass, instance, simple_file):
    """Raise an error if the mv call fails"""
    error = subprocess.CalledProcessError(-1, ["mv"], "test stdout", "test stderr")

    mock_multipass.exec.side_effect = [
        # test call
        mock.Mock(returncode=0),
        # mktemp call
        mock.Mock(stdout="/tmp/mktemp-file\n", returncode=0),
        # chmod call
        None,
        # mv call
        error,
    ]
    mock_multipass.transfer.side_effect = error

    with pytest.raises(MultipassError) as exc_info:
        instance.push_file(
            source=simple_file, destination=pathlib.PurePosixPath("/etc/test.conf")
        )

    assert exc_info.value == MultipassError(
        brief=(
            "Failed to push file '/etc/test.conf' into Multipass instance "
            "'test-instance'."
        ),
        details=errors.details_from_called_process_error(error),
    )


def test_start(mock_multipass, instance):
    instance.start()

    assert mock_multipass.mock_calls == [mock.call.start(instance_name="test-instance")]


def test_stop(mock_multipass, instance):
    instance.stop()

    assert mock_multipass.mock_calls == [
        mock.call.stop(instance_name="test-instance", delay_mins=0)
    ]


def test_stop_all_opts(mock_multipass, instance):
    instance.stop(delay_mins=4)

    assert mock_multipass.mock_calls == [
        mock.call.stop(instance_name="test-instance", delay_mins=4)
    ]


def test_unmount(mock_multipass, instance):
    instance.unmount(target=pathlib.PurePosixPath("/mnt"))

    assert mock_multipass.mock_calls == [mock.call.umount(mount="test-instance:/mnt")]


def test_unmount_all(mock_multipass, instance):
    instance.unmount_all()

    assert mock_multipass.mock_calls == [mock.call.umount(mount="test-instance")]



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""LXD command-line interface helpers."""
import logging
import pathlib
import subprocess
from typing import Optional

import packaging.version

from craft_providers.errors import details_from_called_process_error

from .errors import LXDError

logger = logging.getLogger(__name__)


class LXD:
    """Interface to `lxd` command-line.

    :param lxd_path: Path to lxd.
    :cvar minimum_required_version: Minimum lxd version required for compatibility.
    """

    minimum_required_version = "4.0"

    def __init__(self, *, lxd_path: pathlib.Path = pathlib.Path("lxd")) -> None:
        self.lxd_path = lxd_path

    def init(self, *, auto: bool = False, sudo: bool = False) -> None:
        """Initialize LXD.

        Sudo is required if user is not in lxd group.

        :param auto: Use default settings.
        :param sudo: Use sudo to invoke init.
        """
        cmd = ["sudo"] if sudo else []

        cmd += [str(self.lxd_path), "init"]

        if auto:
            cmd.append("--auto")

        try:
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError as error:
            raise LXDError(
                "Failed to init LXD.",
                details=details_from_called_process_error(error),
            ) from error

    def is_supported_version(self) -> bool:
        """Check if LXD version is supported.

        A helper to check if LXD meets minimum supported version for
        craft-providers (currently >= 4.0).

        :returns: True if installed version is supported.
        """
        minimum_version = packaging.version.parse(self.minimum_required_version)
        version = self.version()

        if "." not in version:
            raise LXDError(
                "Failed to parse LXD version.",
                details=f"Version data returned: {version!r}",
            )

        return packaging.version.parse(version) >= minimum_version

    def version(self) -> str:
        """Query LXD version.

        The version is of the format:
        <major>.<minor>[.<micro>]

        Version examples:
        - 4.13
        - 4.0.5
        - 2.0.12

        :returns: Version string.
        """
        cmd = [str(self.lxd_path), "version"]

        try:
            proc = subprocess.run(cmd, capture_output=True, check=True, text=True)
        except subprocess.CalledProcessError as error:
            raise LXDError(
                "Failed to query LXD version.",
                details=details_from_called_process_error(error),
            ) from error

        return proc.stdout.strip()

    def wait_ready(
        self,
        *,
        sudo: bool = False,
        timeout: Optional[int] = None,
    ) -> None:
        """Wait until LXD is ready.

        Sudo is required if user is not in lxd group.

        :param sudo: Use sudo to invoke waitready.
        :param timeout: Timeout in seconds.
        """
        cmd = ["sudo"] if sudo else []

        cmd += [str(self.lxd_path), "waitready"]

        if timeout is not None:
            cmd.append(f"--timeout={timeout}")

        try:
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError as error:
            raise LXDError(
                "Failed to wait for LXD to get ready.",
                details=details_from_called_process_error(error),
            ) from error



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import pathlib

from craft_providers.util import temp_paths


def test_home_temporary_directory(monkeypatch):
    # Remove the fake home temporary directory fixture set in conftest
    monkeypatch.undo()

    with temp_paths.home_temporary_directory() as tmp_path:
        assert tmp_path.relative_to(pathlib.Path.home())
        assert tmp_path.is_dir() is True

    assert tmp_path.exists() is False


def test_home_temporary_file(monkeypatch):
    # Remove the fake home temporary file fixture set in conftest
    monkeypatch.undo()

    with temp_paths.home_temporary_file() as tmp_file:
        assert tmp_file.relative_to(pathlib.Path.home())
        assert tmp_file.is_file() is True

    assert tmp_file.exists() is False



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import textwrap

import pytest
from craft_providers.util.os_release import parse_os_release


@pytest.mark.parametrize(
    ("config", "expected_dict"),
    [
        (
            textwrap.dedent(
                """\
                NAME="Ubuntu"
                VERSION="22.04 (Jammy Jellyfish)"
                ID=ubuntu
                ID_LIKE=debian
                PRETTY_NAME="Ubuntu 22.04"
                VERSION_ID="22.04"
                HOME_URL="https://www.ubuntu.com/"
                SUPPORT_URL="https://help.ubuntu.com/"
                BUG_REPORT_URL="https://bugs.launchpad.net/ubuntu/"
                PRIVACY_POLICY_URL="https://www.ubuntu.com/legal/terms-and-policies/privacy-policy"
                VERSION_CODENAME=jammy
                UBUNTU_CODENAME=jammy"""
            ),
            {
                "BUG_REPORT_URL": "https://bugs.launchpad.net/ubuntu/",
                "HOME_URL": "https://www.ubuntu.com/",
                "ID": "ubuntu",
                "ID_LIKE": "debian",
                "NAME": "Ubuntu",
                "PRETTY_NAME": "Ubuntu 22.04",
                "PRIVACY_POLICY_URL": "https://www.ubuntu.com/legal/terms-and-policies/privacy-policy",
                "SUPPORT_URL": "https://help.ubuntu.com/",
                "UBUNTU_CODENAME": "jammy",
                "VERSION": "22.04 (Jammy Jellyfish)",
                "VERSION_CODENAME": "jammy",
                "VERSION_ID": "22.04",
            },
        ),
        (
            textwrap.dedent(
                """\
                NAME=Fedora
                VERSION="32 (Workstation Edition)"
                ID=fedora
                VERSION_ID=32
                PRETTY_NAME="Fedora 32 (Workstation Edition)"
                ANSI_COLOR="0;38;2;60;110;180"
                LOGO=fedora-logo-icon
                CPE_NAME="cpe:/o:fedoraproject:fedora:32"
                HOME_URL="https://fedoraproject.org/"
                DOCUMENTATION_URL="https://docs.fedoraproject.org/en-US/fedora/f32/system-administrators-guide/"
                SUPPORT_URL="https://fedoraproject.org/wiki/Communicating_and_getting_help"
                BUG_REPORT_URL="https://bugzilla.redhat.com/"
                REDHAT_BUGZILLA_PRODUCT="Fedora"
                REDHAT_BUGZILLA_PRODUCT_VERSION=32
                REDHAT_SUPPORT_PRODUCT="Fedora"
                REDHAT_SUPPORT_PRODUCT_VERSION=32
                PRIVACY_POLICY_URL="https://fedoraproject.org/wiki/Legal:PrivacyPolicy"
                VARIANT="Workstation Edition"
                VARIANT_ID=workstation
                """
            ),
            {
                "ANSI_COLOR": "0;38;2;60;110;180",
                "BUG_REPORT_URL": "https://bugzilla.redhat.com/",
                "CPE_NAME": "cpe:/o:fedoraproject:fedora:32",
                "DOCUMENTATION_URL": "https://docs.fedoraproject.org/en-US/fedora/f32/system-administrators-guide/",
                "HOME_URL": "https://fedoraproject.org/",
                "ID": "fedora",
                "LOGO": "fedora-logo-icon",
                "NAME": "Fedora",
                "PRETTY_NAME": "Fedora 32 (Workstation Edition)",
                "PRIVACY_POLICY_URL": "https://fedoraproject.org/wiki/Legal:PrivacyPolicy",
                "REDHAT_BUGZILLA_PRODUCT": "Fedora",
                "REDHAT_BUGZILLA_PRODUCT_VERSION": "32",
                "REDHAT_SUPPORT_PRODUCT": "Fedora",
                "REDHAT_SUPPORT_PRODUCT_VERSION": "32",
                "SUPPORT_URL": "https://fedoraproject.org/wiki/Communicating_and_getting_help",
                "VARIANT": "Workstation Edition",
                "VARIANT_ID": "workstation",
                "VERSION": "32 (Workstation Edition)",
                "VERSION_ID": "32",
            },
        ),
        ('foo="double-quotes-to-remove"', {"foo": "double-quotes-to-remove"}),
        ("foo='single-quotes-to-remove'", {"foo": "single-quotes-to-remove"}),
        ('foo="double"-quote-in-the-start', {"foo": '"double"-quote-in-the-start'}),
        ("foo='single'-quote-in-the-start", {"foo": "'single'-quote-in-the-start"}),
        ('foo=double-quote-in-"the"-middle', {"foo": 'double-quote-in-"the"-middle'}),
        ("foo=single-quote-in-'the'-middle", {"foo": "single-quote-in-'the'-middle"}),
        ('foo=double-quote-in-"the-end"', {"foo": 'double-quote-in-"the-end"'}),
        ("foo=single-quote-in-'the-end'", {"foo": "single-quote-in-'the-end'"}),
        ("foo=extra=equal", {"foo": "extra=equal"}),
        ("# ignore commented line", {}),
        ("", {}),
    ],
)
def test_parse_os_release(config, expected_dict):
    assert parse_os_release(config) == expected_dict



#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Helper utilities for craft_providers."""
from __future__ import annotations

import time
from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable

T = TypeVar("T")


def retry_until_timeout(
    timeout: float,
    retry_wait: float,
    func: Callable[[float], T],
    *,
    error: Exception | None = TimeoutError(),
) -> T:
    """Re-run a function until it either succeeds or it times out.

    :param timeout: The length of time (in seconds) before timeout.
    :param retry_wait: The length of time (in seconds) before retrying.
    :param func: The callable. May only take a timeout parameter.
        Must raise an exception on failure, may return anything on success.
    :param error: Exception to raise on timeout or None to pass the error unchanged.
    :returns: The result of the function
    :raises: the passed error from the last exception
    """
    deadline = time.monotonic() + timeout
    soft_deadline = deadline - retry_wait

    while (now := time.monotonic()) < soft_deadline:
        try:
            return func(deadline - now)
        except Exception:
            if time.monotonic() < soft_deadline:
                time.sleep(retry_wait)
    try:
        return func(retry_wait)
    except Exception as exc:
        if error is None:
            raise
        raise error from exc



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import pathlib

import pytest
from craft_providers.util import env_cmd


@pytest.mark.parametrize(
    ("env", "expected"),
    [
        ({}, ["env"]),
        ({"foo": "bar"}, ["env", "foo=bar"]),
        ({"foo": "bar", "foo2": "bar2"}, ["env", "foo=bar", "foo2=bar2"]),
        (
            {"foo": "bar", "foo2": None, "foo3": "baz"},
            ["env", "foo=bar", "-u", "foo2", "foo3=baz"],
        ),
        ({"foo": None}, ["env", "-u", "foo"]),
    ],
)
def test_formulate_command(env, expected):
    assert env_cmd.formulate_command(env) == expected


@pytest.mark.parametrize(
    ("env", "expected"),
    [
        ({}, ["env", "-i"]),
        ({"foo": "bar"}, ["env", "-i", "foo=bar"]),
        ({"foo": "bar", "foo2": "bar2"}, ["env", "-i", "foo=bar", "foo2=bar2"]),
        (
            {"foo": "bar", "foo2": None, "foo3": "baz"},
            ["env", "-i", "foo=bar", "-u", "foo2", "foo3=baz"],
        ),
        ({"foo": None}, ["env", "-i", "-u", "foo"]),
    ],
)
def test_formulate_command_ignore(env, expected):
    assert env_cmd.formulate_command(env, ignore_environment=True) == expected


def test_formulate_command_chdir():
    assert env_cmd.formulate_command(chdir=pathlib.PurePosixPath("/tmp/foo")) == [
        "env",
        "--chdir=/tmp/foo",
    ]


def test_formulate_command_all_opts():
    assert env_cmd.formulate_command(
        env={"VAR_A": "1", "VAR_B": "2"},
        ignore_environment=True,
        chdir=pathlib.PurePosixPath("/tmp/foo"),
    ) == ["env", "--chdir=/tmp/foo", "-i", "VAR_A=1", "VAR_B=2"]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest.mock import call

import pytest
from craft_providers.bases import ubuntu
from craft_providers.errors import BaseConfigurationError
from craft_providers.multipass import MultipassError, MultipassProvider
from craft_providers.multipass.multipass_provider import (
    _BUILD_BASE_TO_MULTIPASS_REMOTE_IMAGE,
    Remote,
    RemoteImage,
)


@pytest.fixture()
def mock_buildd_base_configuration(mocker):
    mock_base_config = mocker.patch(
        "craft_providers.bases.ubuntu.BuilddBase", autospec=True
    )
    mock_base_config.alias = ubuntu.BuilddBaseAlias.JAMMY
    mock_base_config.compatibility_tag = "buildd-base-v2"
    return mock_base_config


@pytest.fixture()
def mock_multipass(mocker):
    return mocker.patch("craft_providers.multipass.Multipass", autospec=True)


@pytest.fixture(autouse=True)
def mock_ensure_multipass_is_ready(mocker):
    return mocker.patch(
        "craft_providers.multipass.multipass_provider.ensure_multipass_is_ready",
        return_value=None,
    )


@pytest.fixture()
def mock_install(mocker):
    return mocker.patch("craft_providers.multipass.multipass_provider.install")


@pytest.fixture(autouse=True)
def mock_is_installed(mocker):
    return mocker.patch(
        "craft_providers.multipass.multipass_provider.is_installed", return_value=True
    )


@pytest.fixture()
def mock_launch(mocker):
    return mocker.patch(
        "craft_providers.multipass.multipass_provider.launch", autospec=True
    )


@pytest.fixture()
def mock_remote_image(mocker):
    """Returns a mock RemoteImage object."""
    _mock_remote_image = mocker.patch(
        "craft_providers.multipass.multipass_provider.RemoteImage", autospec=True
    )
    _mock_remote_image.name = "test-remote:test-image"
    _mock_remote_image.is_stable = True
    return _mock_remote_image


def test_ensure_provider_is_available_installed(
    mock_is_installed, mock_install, mock_ensure_multipass_is_ready
):
    """Verify multipass is installed if it is not already installed."""
    mock_is_installed.return_value = True
    provider = MultipassProvider()

    provider.ensure_provider_is_available()

    mock_is_installed.assert_called_once()
    mock_install.assert_not_called()
    mock_ensure_multipass_is_ready.assert_called_once()


def test_ensure_provider_is_available_not_installed(
    mock_is_installed, mock_install, mock_ensure_multipass_is_ready
):
    """Verify multipass is not re-installed if it is already installed."""
    mock_is_installed.return_value = False
    provider = MultipassProvider()

    provider.ensure_provider_is_available()

    mock_is_installed.assert_called_once()
    mock_install.assert_called_once()
    mock_ensure_multipass_is_ready.assert_called_once()


@pytest.mark.parametrize("is_installed", [True, False])
def test_is_provider_installed(is_installed, mock_is_installed):
    """Verify `is_provider_installed` checks if Multipass is actually installed."""
    mock_is_installed.return_value = is_installed
    provider = MultipassProvider()

    assert provider.is_provider_installed() == is_installed


def test_create_environment(mocker):
    mock_multipass_instance = mocker.patch(
        "craft_providers.multipass.multipass_provider.MultipassInstance"
    )

    provider = MultipassProvider()
    provider.create_environment(instance_name="test-name")

    mock_multipass_instance.assert_called_once_with(name="test-name")


@pytest.mark.parametrize(
    ("build_base", "remote_image"), _BUILD_BASE_TO_MULTIPASS_REMOTE_IMAGE.items()
)
def test_launched_environment(
    build_base,
    remote_image,
    mock_launch,
    tmp_path,
):
    """Verify `launched_environment()` function."""
    provider = MultipassProvider()
    base_configuration = ubuntu.BuilddBase(alias=build_base)
    with provider.launched_environment(
        project_name="test-project",
        project_path=tmp_path,
        base_configuration=base_configuration,
        instance_name="test-instance-name",
        allow_unstable=True,
    ) as instance:
        assert instance is not None
        assert mock_launch.mock_calls == [
            call(
                name="test-instance-name",
                base_configuration=base_configuration,
                image_name=remote_image.name,
                cpus=2,
                disk_gb=64,
                mem_gb=2,
                auto_clean=True,
            ),
        ]
        mock_launch.reset_mock()

    assert mock_launch.mock_calls == [
        call().unmount_all(),
        call().stop(),
    ]


@pytest.mark.parametrize(
    ("is_stable", "allow_unstable"),
    [
        # unstable images can only be launched when `allow_unstable=True`
        (False, True),
        # stable images can be launched regardless of `allow_unstable`
        (True, False),
        (True, True),
    ],
)
def test_launched_environment_stable(
    is_stable,
    allow_unstable,
    mock_buildd_base_configuration,
    mock_launch,
    mock_remote_image,
    mocker,
    tmp_path,
):
    """Verify allow_unstable parameter works as expected."""
    mock_remote_image.is_stable = is_stable
    mocker.patch(
        "craft_providers.multipass.multipass_provider._get_remote_image",
        return_value=mock_remote_image,
    )

    provider = MultipassProvider()
    with provider.launched_environment(
        project_name="test-project",
        project_path=tmp_path,
        base_configuration=mock_buildd_base_configuration,
        instance_name="test-instance-name",
        allow_unstable=allow_unstable,
    ) as instance:
        assert instance is not None
        assert mock_launch.mock_calls == [
            call(
                name="test-instance-name",
                base_configuration=mock_buildd_base_configuration,
                image_name="test-remote:test-image",
                cpus=2,
                disk_gb=64,
                mem_gb=2,
                auto_clean=True,
            ),
        ]
        mock_launch.reset_mock()

    assert mock_launch.mock_calls == [
        call().unmount_all(),
        call().stop(),
    ]


def test_launched_environment_unstable_image_error(
    mock_buildd_base_configuration,
    mock_launch,
    mock_remote_image,
    mocker,
    tmp_path,
):
    """Raise an error when `allow_unstable=False` and image is unstable."""
    mock_remote_image.is_stable = False
    mocker.patch(
        "craft_providers.multipass.multipass_provider._get_remote_image",
        return_value=mock_remote_image,
    )

    provider = MultipassProvider()
    with pytest.raises(MultipassError) as raised, provider.launched_environment(
        project_name="test-project",
        project_path=tmp_path,
        base_configuration=mock_buildd_base_configuration,
        instance_name="test-instance-name",
    ):
        pass

    assert raised.value == MultipassError(
        brief="Cannot launch unstable image 'test-remote:test-image'.",
        details=(
            "Devel or daily images are not guaranteed and are intended for "
            "experimental use only."
        ),
        resolution=(
            "Set parameter `allow_unstable` to True to launch unstable images."
        ),
    )


def test_launched_environment_launch_base_configuration_error(mock_launch, tmp_path):
    error = BaseConfigurationError(brief="fail")
    mock_launch.side_effect = error
    provider = MultipassProvider()
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(MultipassError, match="fail") as raised:
        with provider.launched_environment(
            project_name="test-project",
            project_path=tmp_path,
            base_configuration=base_configuration,
            instance_name="test-instance-name",
        ):
            pass

    assert raised.value.__cause__ is error


@pytest.mark.parametrize("remote", Remote)
def test_remote_image_name(remote):
    """Verify RemoteImage name is formulated properly."""
    remote_image = RemoteImage(remote=remote, image_name="test-name")
    assert remote_image.name == f"{remote.value}:test-name"


@pytest.mark.parametrize(
    ("remote_image", "is_stable"),
    [
        # 'release' and 'snapcraft' remotes are stable
        (RemoteImage(remote=Remote.RELEASE, image_name="test-name"), True),
        (RemoteImage(remote=Remote.SNAPCRAFT, image_name="test-name"), True),
        # 'daily' remote is not stable
        (RemoteImage(remote=Remote.DAILY, image_name="test-name"), False),
        # 'devel' images are not stable, regardless of remote
        (RemoteImage(remote=Remote.RELEASE, image_name="devel"), False),
        (RemoteImage(remote=Remote.DAILY, image_name="devel"), False),
        (RemoteImage(remote=Remote.SNAPCRAFT, image_name="devel"), False),
    ],
)
def test_remote_is_stable(remote_image, is_stable):
    """RemoteImages should be properly marked as stable or unstable."""
    assert remote_image.is_stable == is_stable


def test_multipass_name():
    """Verify MultipassProvider's name."""
    provider = MultipassProvider()

    assert provider.name == "Multipass"


def test_multipass_install_recommendation():
    """Verify MultipassProvider's install recommendation."""
    provider = MultipassProvider()

    assert (
        provider.install_recommendation
        == "Visit https://multipass.run/install for instructions to install Multipass."
    )



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Ubuntu image(s)."""
import enum
import io
import logging
import pathlib
import subprocess
from textwrap import dedent
from typing import Dict, List, Optional

from craft_providers.actions.snap_installer import Snap
from craft_providers.base import Base
from craft_providers.errors import (
    BaseCompatibilityError,
    BaseConfigurationError,
    details_from_called_process_error,
)
from craft_providers.executor import Executor

logger = logging.getLogger(__name__)


class BuilddBaseAlias(enum.Enum):
    """Mappings for supported buildd images."""

    XENIAL = "16.04"
    BIONIC = "18.04"
    FOCAL = "20.04"
    JAMMY = "22.04"
    LUNAR = "23.04"
    DEVEL = "devel"


class BuilddBase(Base):
    """Support for Ubuntu minimal buildd images.

    :cvar compatibility_tag: Tag/Version for variant of build configuration and
        setup.  Any change to this version would indicate that prior [versioned]
        instances are incompatible and must be cleaned.  As such, any new value
        should be unique to old values (e.g. incrementing).  It is suggested to
        extend this tag, not overwrite it, e.g.: compatibility_tag =
        f"{appname}-{BuildBase.compatibility_tag}.{apprevision}" to ensure base
        compatibility levels are maintained.
    :cvar instance_config_path: Path to persistent environment configuration
        used for compatibility checks (or other data).  Set to
        /etc/craft-instance.conf, but may be overridden for application-specific
        reasons.
    :cvar instance_config_class: Class defining instance configuration.  May be
        overridden with an application-specific subclass of InstanceConfiguration
        to enable application-specific extensions.

    :param alias: Base alias / version.
    :param environment: Environment to set in /etc/environment.
    :param hostname: Hostname to configure.
    :param snaps: Optional list of snaps to install on the base image.
    :param packages: Optional list of system packages to install on the base image.
    :param use_default_packages: Optional bool to enable/disable default packages.
    :param cache_path: Optional path to the shared cache directory. If this is
        provided, shared cache directories will be mounted as appropriate.
    """

    compatibility_tag: str = f"buildd-{Base.compatibility_tag}"

    def __init__(
        self,
        *,
        alias: BuilddBaseAlias,
        compatibility_tag: Optional[str] = None,
        environment: Optional[Dict[str, Optional[str]]] = None,
        hostname: str = "craft-buildd-instance",
        snaps: Optional[List[Snap]] = None,
        packages: Optional[List[str]] = None,
        use_default_packages: bool = True,
        cache_path: Optional[pathlib.Path] = None,
    ) -> None:
        self.alias: BuilddBaseAlias = alias
        self._cache_path = cache_path

        if environment is None:
            self._environment = self.default_command_environment()
        else:
            self._environment = environment

        if compatibility_tag:
            self.compatibility_tag = compatibility_tag

        self._set_hostname(hostname)

        self._packages: List[str] = []
        if use_default_packages:
            self._packages.extend(
                [
                    "apt-utils",
                    "build-essential",
                    "curl",
                    "fuse",
                    "udev",
                    "python3",
                    "python3-dev",
                    "python3-pip",
                    "python3-wheel",
                    "python3-setuptools",
                ]
            )

        if packages:
            self._packages.extend(packages)

        self._snaps = snaps

    def _disable_automatic_apt(self, executor: Executor) -> None:
        """Disable automatic apt actions.

        This should happen as soon as possible in the instance overall setup,
        to reduce the chances of an automatic apt work being triggered during
        the setup itself (because it includes apt work which may clash
        the triggered unattended jobs).
        """
        # set the verification frequency in 10000 days and disable the upgrade
        content = dedent(
            """\
            APT::Periodic::Update-Package-Lists "10000";
            APT::Periodic::Unattended-Upgrade "0";
        """
        ).encode()
        executor.push_file_io(
            destination=pathlib.PurePosixPath("/etc/apt/apt.conf.d/20auto-upgrades"),
            content=io.BytesIO(content),
            file_mode="0644",
        )

    def _ensure_os_compatible(self, executor: Executor) -> None:
        """Ensure OS is compatible with Base."""
        os_release = self._get_os_release(executor=executor)

        os_name = os_release.get("NAME")
        if os_name != "Ubuntu":
            raise BaseCompatibilityError(
                reason=f"Expected OS 'Ubuntu', found {os_name!r}"
            )

        compat_version_id = self.alias.value
        version_id = os_release.get("VERSION_ID")

        if compat_version_id == BuilddBaseAlias.DEVEL.value:
            logger.debug(
                "Ignoring OS version mismatch for %r because base is %r.",
                version_id,
                compat_version_id,
            )
            return

        if version_id != compat_version_id:
            raise BaseCompatibilityError(
                reason=(
                    f"Expected OS version {compat_version_id!r},"
                    f" found {version_id!r}"
                )
            )

    def _update_apt_sources(self, executor: Executor, codename: str) -> None:
        """Update the codename in the apt source config files.

        :param codename: New codename to use in apt source config files (i.e. 'lunar')
        """
        apt_source = "/etc/apt/sources.list"
        apt_source_dir = "/etc/apt/sources.list.d/"
        cloud_config = "/etc/cloud/cloud.cfg"

        # get the current ubuntu codename
        os_release = self._get_os_release(executor=executor)
        version_codename = os_release.get("VERSION_CODENAME")
        logger.debug("Updating apt sources from %r to %r.", version_codename, codename)

        # replace all occurrences of the codename in the `sources.list` file
        sed_command = ["sed", "-i", f"s/{version_codename}/{codename}/g"]
        try:
            self._execute_run(
                [*sed_command, apt_source],
                executor=executor,
                timeout=self._timeout_simple,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief=f"Failed to update {apt_source!r}.",
                details=details_from_called_process_error(error),
            ) from error

        # if cloud-init and cloud.cfg isn't present, then raise an error
        try:
            self._execute_run(
                ["test", "-s", cloud_config],
                executor=executor,
                timeout=self._timeout_simple,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief=(
                    f"Could not update {cloud_config!r} because it is empty or "
                    "does not exist."
                ),
                details=details_from_called_process_error(error),
            ) from error

        # update cloud.cfg to prevent the sources.list file from being reset
        logger.debug("Updating %r to preserve apt sources.", cloud_config)
        try:
            self._execute_run(
                # 'aapt' is not a typo, the first 'a' is the sed command to append
                # this is a shlex-compatible way to append to a file
                ["sed", "-i", "$ aapt_preserve_sources_list: true", cloud_config],
                executor=executor,
                timeout=self._timeout_simple,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief=f"Failed to update {cloud_config!r}.",
                details=details_from_called_process_error(error),
            ) from error

        # running `find` and `sed` as two separate calls may appear unoptimized,
        # but these shell commands will pass through `shlex.join()` before being
        # executed, which means one-liners like `find -exec sed` or
        # `find | xargs sed` cannot be used

        try:
            additional_source_files = self._execute_run(
                ["find", apt_source_dir, "-type", "f", "-name", "*.list"],
                executor=executor,
                text=True,
                timeout=self._timeout_simple,
            ).stdout.strip()
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief=f"Failed to find apt source files in {apt_source_dir!r}.",
                details=details_from_called_process_error(error),
            ) from error

        # if there are config files in `sources.list.d/`, then update them
        if additional_source_files:
            try:
                self._execute_run(
                    [*sed_command, apt_source_dir + "*.list"],
                    executor=executor,
                    timeout=self._timeout_simple,
                )
            except subprocess.CalledProcessError as error:
                raise BaseConfigurationError(
                    brief=f"Failed to update apt source files in {apt_source_dir!r}.",
                    details=details_from_called_process_error(error),
                ) from error

    def _post_setup_os(self, executor: Executor) -> None:
        """Ubuntu specific post-setup OS tasks."""
        self._disable_automatic_apt(executor=executor)

    def _setup_network(self, executor: Executor) -> None:
        """Set up the basic network with systemd-networkd and systemd-resolved."""
        self._setup_hostname(executor=executor)
        self._setup_resolved(executor=executor)
        self._setup_networkd(executor=executor)

    def _pre_setup_packages(self, executor: Executor) -> None:
        """Configure apt, update database."""
        executor.push_file_io(
            destination=pathlib.Path("/etc/apt/apt.conf.d/00no-recommends"),
            content=io.BytesIO(b'APT::Install-Recommends "false";\n'),
            file_mode="0644",
        )

        executor.push_file_io(
            destination=pathlib.Path("/etc/apt/apt.conf.d/00update-errors"),
            content=io.BytesIO(b'APT::Update::Error-Mode "any";\n'),
            file_mode="0644",
        )

        # devel images should use the devel repository
        if self.alias == BuilddBaseAlias.DEVEL:
            self._update_apt_sources(
                executor=executor,
                codename=BuilddBaseAlias.DEVEL.value,
            )

        try:
            self._execute_run(
                ["apt-get", "update"],
                executor=executor,
                verify_network=True,
                timeout=self._timeout_unpredictable,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to update apt cache.",
                details=details_from_called_process_error(error),
            ) from error

    def _setup_packages(self, executor: Executor) -> None:
        """Use apt install required packages and user-defined packages."""
        if not self._packages:
            return
        try:
            command = ["apt-get", "install", "-y", *self._packages]
            self._execute_run(
                command,
                executor=executor,
                verify_network=True,
                timeout=self._timeout_unpredictable,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to install packages.",
                details=details_from_called_process_error(error),
            ) from error

    def _setup_snapd(self, executor: Executor) -> None:
        """Install snapd and dependencies and wait until ready."""
        try:
            self._execute_run(
                ["apt-get", "install", "-y", "snapd"],
                executor=executor,
                verify_network=True,
                timeout=self._timeout_complex,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to setup snapd.",
                details=details_from_called_process_error(error),
            ) from error

    def _clean_up(self, executor: Executor) -> None:
        self._execute_run(
            ["apt-get", "autoremove", "-y"],
            executor=executor,
            timeout=self._timeout_complex,
        )
        self._execute_run(
            ["apt-get", "clean", "-y"],
            executor=executor,
            timeout=self._timeout_complex,
        )


# Backward compatible, will be removed in 2.0
default_command_environment = BuilddBase.default_command_environment



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Abstract base class for Providers.

The Provider, LXDProvider, and MultipassProvider classes are not stable and are
likely to change. These classes will be stable and recommended for use in the release
of craft-providers 2.0.
"""

import contextlib
import logging
import pathlib
from abc import ABC, abstractmethod
from typing import Generator, Optional

from .base import Base
from .executor import Executor

logger = logging.getLogger(__name__)


class Provider(ABC):
    """Build environment provider."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the provider."""

    @property
    @abstractmethod
    def install_recommendation(self) -> str:
        """Recommended way to install the provider."""

    def clean_project_environments(self, *, instance_name: str) -> None:
        """Clean the provider environment.

        :param instance_name: name of the instance to clean
        """
        # Nothing to do if provider is not installed.
        if not self.is_provider_installed():
            logger.debug(
                "Not cleaning environment because the provider is not installed."
            )
            return

        environment = self.create_environment(instance_name=instance_name)
        if environment.exists():
            environment.delete()

    @classmethod
    @abstractmethod
    def ensure_provider_is_available(cls) -> None:
        """Ensure provider is available, prompting the user to install it if required.

        :raises ProviderError: if provider is not available.
        """

    @classmethod
    @abstractmethod
    def is_provider_installed(cls) -> bool:
        """Check if provider is installed.

        :returns: True if installed.
        """

    @abstractmethod
    def create_environment(self, *, instance_name: str) -> Executor:
        """Create a bare environment for specified base.

        No initializing, launching, or cleaning up of the environment occurs.

        :param instance_name: name of the instance to create
        """

    @abstractmethod
    @contextlib.contextmanager
    def launched_environment(
        self,
        *,
        project_name: str,
        project_path: pathlib.Path,
        base_configuration: Base,
        build_base: Optional[str] = None,
        instance_name: str,
        allow_unstable: bool = False,
    ) -> Generator[Executor, None, None]:
        """Configure and launch environment for specified base.

        When this method loses context, all directories are unmounted and the
        environment is stopped. For more control of environment setup and teardown,
        use `create_environment()` instead.

        :param project_name: Name of project.
        :param project_path: Path to project.
        :param base_configuration: Base configuration to apply to instance.
        :param build_base: Base to build from. (Deprecated)
        :param instance_name: Name of the instance to launch.
        :param allow_unstable: If true, allow unstable images to be launched.
        """



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""LXD Provider class."""

import contextlib
import logging
import pathlib
from datetime import timedelta
from typing import Iterator, Optional

from craft_providers import Executor, Provider
from craft_providers.base import Base
from craft_providers.errors import BaseConfigurationError

from .errors import LXDError, LXDUnstableImageError
from .installer import ensure_lxd_is_ready, install, is_installed
from .launcher import launch
from .lxc import LXC
from .lxd_instance import LXDInstance
from .remotes import get_remote_image

logger = logging.getLogger(__name__)


class LXDProvider(Provider):
    """LXD build environment provider.

    This class is not stable and is likely to change. This class will be stable and
    recommended for use in the release of craft-providers 2.0.

    :param lxc: Optional lxc client to use.
    :param lxd_project: LXD project to use (default is default).
    :param lxd_remote: LXD remote to use (default is local).
    """

    def __init__(
        self,
        *,
        lxc: LXC = LXC(),
        lxd_project: str = "default",
        lxd_remote: str = "local",
    ) -> None:
        self.lxc = lxc
        self.lxd_project = lxd_project
        self.lxd_remote = lxd_remote

    @property
    def name(self) -> str:
        """Name of the provider."""
        return "LXD"

    @property
    def install_recommendation(self) -> str:
        """Recommended way to install the provider."""
        return "Visit https://ubuntu.com/lxd/install for instructions to install LXD."

    @classmethod
    def ensure_provider_is_available(cls) -> None:
        """Ensure provider is available and ready, installing if required.

        :raises LXDInstallationError: if LXD cannot be installed
        :raises LXDError: if provider is not available
        """
        if not is_installed():
            install()
        ensure_lxd_is_ready()

    @classmethod
    def is_provider_installed(cls) -> bool:
        """Check if provider is installed.

        :returns: True if installed.
        """
        return is_installed()

    def create_environment(self, *, instance_name: str) -> Executor:
        """Create a bare environment for specified base.

        No initializing, launching, or cleaning up of the environment occurs.

        :param instance_name: Name of the instance.
        """
        return LXDInstance(
            name=instance_name,
            project=self.lxd_project,
            remote=self.lxd_remote,
        )

    @contextlib.contextmanager
    def launched_environment(
        self,
        *,
        project_name: str,
        project_path: pathlib.Path,
        base_configuration: Base,
        build_base: Optional[str] = None,
        instance_name: str,
        allow_unstable: bool = False,
    ) -> Iterator[Executor]:
        """Configure and launch environment for specified base.

        When this method loses context, all directories are unmounted and the
        environment is stopped. For more control of environment setup and teardown,
        use `create_environment()` instead.

        :param project_name: Name of project.
        :param project_path: Path to project.
        :param base_configuration: Base configuration to apply to instance.
        :param build_base: Base to build from. (Deprecated)
        :param instance_name: Name of the instance to launch.
        :param allow_unstable: If true, allow unstable images to be launched.

        :raises LXDError: if instance cannot be configured and launched.
        """
        if build_base:
            logger.warning(
                "Deprecated: Parameter 'build_base' is deprecated and should "
                "not be used. The build base now comes from the "
                "base_configuration's alias."
            )

        image = get_remote_image(base_configuration)
        image.add_remote(lxc=self.lxc)

        # only allow launching unstable images when opted-in with `allow_unstable`
        if not image.is_stable and not allow_unstable:
            raise LXDUnstableImageError(
                brief=(
                    f"Cannot launch an unstable image {image.image_name!r} from remote "
                    f"{image.remote_name!r}"
                ),
            )

        # unstable images should be refreshed more often
        expiration = timedelta(days=90) if image.is_stable else timedelta(days=14)

        try:
            instance = launch(
                name=instance_name,
                base_configuration=base_configuration,
                image_name=image.image_name,
                image_remote=image.remote_name,
                auto_clean=True,
                auto_create_project=True,
                map_user_uid=True,
                uid=project_path.stat().st_uid,
                use_base_instance=True,
                project=self.lxd_project,
                remote=self.lxd_remote,
                expiration=expiration,
            )
        except BaseConfigurationError as error:
            raise LXDError(str(error)) from error

        try:
            yield instance
        finally:
            # Ensure to unmount everything and stop instance upon completion.
            instance.unmount_all()
            instance.stop()



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

from unittest import mock

import pytest
from craft_providers import lxd
from craft_providers.lxd import project


@pytest.fixture(autouse=True)
def mock_lxc():
    with mock.patch("craft_providers.lxd.LXC", spec=lxd.LXC) as lxc_mock:
        lxc_mock.return_value.profile_show.return_value = {"test": "config"}
        yield lxc_mock.return_value


def test_create_with_default_profile(mock_lxc):
    project.create_with_default_profile(
        lxc=mock_lxc,
        project="test-project",
        profile="test-profile",
        profile_project="test-profile-project",
        remote="test-remote",
    )

    assert mock_lxc.mock_calls == [
        mock.call.project_create(project="test-project", remote="test-remote"),
        mock.call.profile_show(
            profile="test-profile", project="test-profile-project", remote="test-remote"
        ),
        mock.call.profile_edit(
            profile="default",
            project="test-project",
            config={"test": "config"},
            remote="test-remote",
        ),
    ]


def test_purge(mock_lxc):
    mock_lxc.project_list.return_value = ["test-project", "default"]
    mock_lxc.list_names.return_value = ["test-instance1", "test-instance2"]
    mock_lxc.image_list.return_value = [{"fingerprint": "i1"}, {"fingerprint": "i2"}]

    project.purge(lxc=mock_lxc, project="test-project", remote="test-remote")

    assert mock_lxc.mock_calls == [
        mock.call.project_list(remote="test-remote"),
        mock.call.list_names(project="test-project", remote="test-remote"),
        mock.call.delete(
            instance_name="test-instance1",
            project="test-project",
            remote="test-remote",
            force=True,
        ),
        mock.call.delete(
            instance_name="test-instance2",
            project="test-project",
            remote="test-remote",
            force=True,
        ),
        mock.call.image_list(project="test-project"),
        mock.call.image_delete(
            image="i1", project="test-project", remote="test-remote"
        ),
        mock.call.image_delete(
            image="i2", project="test-project", remote="test-remote"
        ),
        mock.call.project_delete(project="test-project", remote="test-remote"),
    ]


def test_purge_no_project(mock_lxc):
    mock_lxc.project_list.return_value = ["default"]
    mock_lxc.list.return_value = ["test-instance1", "test-instance2"]
    mock_lxc.image_list.return_value = [{"fingerprint": "i1"}, {"fingerprint": "i2"}]

    project.purge(lxc=mock_lxc, project="test-project", remote="test-remote")

    assert mock_lxc.mock_calls == [
        mock.call.project_list(remote="test-remote"),
    ]



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""LXD Instance Executor."""

import hashlib
import io
import logging
import os
import pathlib
import re
import shutil
import subprocess
import tempfile
from typing import Any, Dict, List, Optional

from craft_providers.const import TIMEOUT_SIMPLE
from craft_providers.errors import details_from_called_process_error
from craft_providers.executor import Executor
from craft_providers.lxd.errors import LXDError
from craft_providers.lxd.lxc import LXC
from craft_providers.util import env_cmd

logger = logging.getLogger(__name__)


class LXDInstance(Executor):
    """LXD Instance Lifecycle."""

    def __init__(
        self,
        *,
        name: str,
        default_command_environment: Optional[Dict[str, Optional[str]]] = None,
        project: str = "default",
        remote: str = "local",
        lxc: Optional[LXC] = None,
    ) -> None:
        """Create an LXD executor.

        To comply with LXD naming conventions, the supplied name is converted to a
        LXD-compatible name before creating the instance.

        :param name: Unique name of the lxd instance
        :param default_command_environment: command environment
        :param project: name of lxd project
        :param remote: name of lxd remote
        :param lxc: LXC instance object
        """
        super().__init__()

        if default_command_environment is not None:
            self.default_command_environment = default_command_environment
        else:
            self.default_command_environment = {}

        self.name = name
        self._set_instance_name()
        self.project = project
        self.remote = remote

        if lxc is None:
            self.lxc = LXC()
        else:
            self.lxc = lxc

    def _set_instance_name(self) -> None:
        """Convert a name to a LXD-compatible name.

        LXD naming convention:
        - between 1 and 63 characters long
        - made up exclusively of letters, numbers, and hyphens from the ASCII table
        - not begin with a digit or a hyphen
        - not end with a hyphen

        To create a LXD-compatible name, invalid characters are removed, the name is
        truncated to 40 characters, then a hash is appended:
        <truncated-name>-<hash-of-name>
        └     1 - 40   ┘1└     20     ┘

        :param name: name of instance
        :raises LXDError: if name contains no alphanumeric characters
        """
        # remove anything that is not an alphanumeric characters or hyphen
        name_with_valid_chars = re.sub(r"[^\w-]", "", self.name)
        if not name_with_valid_chars:
            raise LXDError(
                brief=f"failed to create LXD instance with name {self.name!r}.",
                details="name must contain at least one alphanumeric character",
            )

        # trim digits and hyphens from the beginning and hyphens from the end
        trimmed_name = re.compile(r"^[0-9-]*(?P<valid_name>.*?)[-]*$").search(
            name_with_valid_chars
        )
        if not trimmed_name or not trimmed_name.group("valid_name"):
            raise LXDError(
                brief=f"failed to create LXD instance with name {self.name!r}.",
                details="name must contain at least one alphanumeric character",
            )
        valid_name = trimmed_name.group("valid_name")

        # if the original name meets LXD's naming convention, then use the original name
        if self.name == valid_name and len(self.name) <= 63:
            instance_name = self.name

        # else, continue converting the name
        else:
            # truncate to 40 characters
            truncated_name = valid_name[:40]
            # hash the entire name, not the truncated name
            hashed_name = hashlib.sha1(self.name.encode()).hexdigest()[:20]
            instance_name = f"{truncated_name}-{hashed_name}"

        self.instance_name = instance_name
        logger.debug("Set LXD instance name to %r", instance_name)

    def _finalize_lxc_command(
        self,
        command: List[str],
        *,
        env: Optional[Dict[str, Optional[str]]] = None,
    ) -> List[str]:
        """Wrap a command to run as root with specified environment.

        LXD will run commands as root.

        Account for the command environment by using the default command
        environment as the baseline, updating it to reflect the command's env
        parameter, if any.

        :param command: Command to execute.
        :param env: Additional environment flags to set/unset.

        :returns: List of command strings for multipass exec.
        """
        command_env = self.default_command_environment.copy()

        if env:
            command_env.update(env)

        if command_env:
            return env_cmd.formulate_command(command_env) + command

        return command

    def push_file_io(
        self,
        *,
        destination: pathlib.PurePath,
        content: io.BytesIO,
        file_mode: str,
        group: str = "root",
        user: str = "root",
    ) -> None:
        """Create or replace file with content and file mode.

        :param destination: Path to file.
        :param content: Contents of file.
        :param file_mode: File mode string (e.g. '0644').
        :param group: File group owner/id.
        :param user: File user owner/id.

        :raises LXDError: On unexpected error.
        """
        with tempfile.NamedTemporaryFile() as temp_file:
            shutil.copyfileobj(content, temp_file)  # type: ignore # mypy #15031
            # Ensure the file is written to disk.
            temp_file.flush()

            temp_path = pathlib.Path(temp_file.name)
            self.lxc.file_push(
                instance_name=self.instance_name,
                source=temp_path,
                destination=destination,
                mode=file_mode,
                project=self.project,
                remote=self.remote,
            )

            # We don't use gid/uid for file_push() in case we don't know the
            # user/group IDs in advance.  Just chown it.
            try:
                self.execute_run(
                    ["chown", f"{user}:{group}", destination.as_posix()],
                    capture_output=True,
                    check=True,
                    timeout=TIMEOUT_SIMPLE,
                )
            except subprocess.CalledProcessError as error:
                raise LXDError(
                    brief=(
                        f"Failed to create file {destination.as_posix()!r}"
                        f" in instance {self.instance_name!r}."
                    ),
                    details=details_from_called_process_error(error),
                ) from error

    def delete(self, force: bool = True) -> None:
        """Delete instance.

        :param force: Delete even if running.

        :raises LXDError: On unexpected error.
        """
        return self.lxc.delete(
            instance_name=self.instance_name,
            project=self.project,
            remote=self.remote,
            force=force,
        )

    def execute_popen(
        self,
        command: List[str],
        *,
        cwd: Optional[pathlib.PurePath] = None,
        env: Optional[Dict[str, Optional[str]]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> subprocess.Popen:
        """Execute a command in instance, using subprocess.Popen().

        The process' environment will inherit the execution environment's
        default environment (PATH, etc.), but can be additionally configured via
        env parameter.

        :param command: Command to execute.
        :param cwd: Working directory for the process inside the instance.
        :param env: Additional environment to set for process.
        :param timeout: Timeout (in seconds) for the command.
        :param kwargs: Additional keyword arguments to pass.

        :returns: Popen instance.
        """
        cwd_path = None if cwd is None else cwd.as_posix()

        return self.lxc.exec(
            instance_name=self.instance_name,
            command=self._finalize_lxc_command(command=command, env=env),
            project=self.project,
            remote=self.remote,
            runner=subprocess.Popen,
            timeout=timeout,
            cwd=cwd_path,
            **kwargs,
        )

    def execute_run(
        self,
        command: List[str],
        *,
        cwd: Optional[pathlib.PurePath] = None,
        env: Optional[Dict[str, Optional[str]]] = None,
        timeout: Optional[float] = None,
        check: bool = False,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        """Execute a command using subprocess.run().

        The process' environment will inherit the execution environment's
        default environment (PATH, etc.), but can be additionally configured via
        env parameter.

        :param command: Command to execute.
        :param cwd: Working directory for the process inside the instance.
        :param env: Additional environment to set for process.
        :param timeout: Timeout (in seconds) for the command.
        :param check: Raise an exception if the command fails.
        :param kwargs: Keyword args to pass to subprocess.run().

        :returns: Completed process.

        :raises subprocess.CalledProcessError: if command fails and check is
            True.
        """
        cwd_path = None if cwd is None else cwd.as_posix()

        return self.lxc.exec(
            instance_name=self.instance_name,
            command=self._finalize_lxc_command(command=command, env=env),
            project=self.project,
            remote=self.remote,
            runner=subprocess.run,
            timeout=timeout,
            cwd=cwd_path,
            check=check,
            **kwargs,
        )

    def exists(self) -> bool:
        """Check if instance exists.

        :returns: True if instance exists.

        :raises LXDError: On unexpected error.
        """
        return self._get_state() is not None

    def _get_disk_devices(self) -> Dict[str, Any]:
        """Query instance and return dictionary of disk devices."""
        devices = self.lxc.config_device_show(
            instance_name=self.instance_name, project=self.project, remote=self.remote
        )

        disks = {}
        for name, config in devices.items():
            if config.get("type") == "disk":
                disks[name] = config

                # Ensure the expected keys are in config.
                if "path" not in config or "source" not in config:
                    raise LXDError(
                        brief=f"Failed to parse lxc device {name!r}.",
                        details=f"* Device configuration: {devices!r}",
                    )

        return disks

    def _get_state(self) -> Optional[Dict[str, Any]]:
        """Get state configuration for instance.

        :returns: State information parsed from lxc if instance exists, else
                  None.

        :raises LXDError: On unexpected error.
        """
        instances = self.lxc.list(project=self.project, remote=self.remote)

        for instance in instances:
            if instance["name"] == self.instance_name:
                return instance

        return None

    def is_mounted(
        self, *, host_source: pathlib.Path, target: pathlib.PurePath
    ) -> bool:
        """Check if path is mounted at target.

        :param host_source: Host path to check.
        :param target: Instance path to check.

        :returns: True if host_source is mounted at target.

        :raises LXDError: On unexpected error.
        """
        disks = self._get_disk_devices()

        return any(
            disk["path"] == target.as_posix()
            and disk["source"] == host_source.as_posix()
            for _, disk in disks.items()
        )

    def is_running(self) -> bool:
        """Check if instance is running.

        :returns: True if instance is running.

        :raises LXDError: On unexpected error.
        """
        state = self._get_state()
        if state is None:
            raise LXDError(brief=f"Instance {self.instance_name!r} does not exist.")

        return state.get("status") == "Running"

    def launch(
        self,
        *,
        image: str,
        image_remote: str,
        map_user_uid: bool = False,
        ephemeral: bool = False,
        uid: Optional[int] = None,
    ) -> None:
        """Launch instance.

        :param image: Image name to launch.
        :param image_remote: Image remote name.
        :param map_user_uid: Whether id mapping should be used.
        :param uid: If ``map_user_uid`` is True,
                    the host user ID to map to instance root.
        :param ephemeral: Flag to enable ephemeral instance.

        :raises LXDError: On unexpected error.
        """
        config_keys = {}

        if map_user_uid:
            if not uid:
                uid = os.getuid()
            config_keys["raw.idmap"] = f"both {uid!s} 0"

        if self._host_supports_mknod():
            config_keys["security.syscalls.intercept.mknod"] = "true"

        self.lxc.launch(
            config_keys=config_keys,
            ephemeral=ephemeral,
            instance_name=self.instance_name,
            image=image,
            image_remote=image_remote,
            project=self.project,
            remote=self.remote,
        )

    def mount(self, *, host_source: pathlib.Path, target: pathlib.PurePath) -> None:
        """Mount host source directory to target mount point.

        Checks first to see if already mounted.
        The source will be mounted as a disk named "disk-{target.as_posix()}".

        :param host_source: Host path to mount.
        :param target: Instance path to mount to.

        :raises LXDError: On unexpected error.
        """
        if self.is_mounted(host_source=host_source, target=target):
            return

        self.lxc.config_device_add_disk(
            instance_name=self.instance_name,
            source=host_source,
            path=target,
            device=f"disk-{target.as_posix()}",
            project=self.project,
            remote=self.remote,
        )

    def _host_supports_mknod(self) -> bool:
        """Check if host supports mknod in container.

        See: https://documentation.ubuntu.com/lxd/en/latest/syscall-interception/

        :returns: True if mknod is supported.

        :raises LXDError: On unexpected error.
        """
        cfg = self.lxc.info(project=self.project, remote=self.remote)
        env = cfg.get("environment", {})
        kernel_features = env.get("kernel_features", {})
        seccomp_listener = kernel_features.get("seccomp_listener", "false")

        return seccomp_listener == "true"

    def pull_file(self, *, source: pathlib.PurePath, destination: pathlib.Path) -> None:
        """Copy a file from the environment to host.

        :param source: Environment file to copy.
        :param destination: Host file path to copy to.  Parent directory
            (destination.parent) must exist.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist.
        :raises LXDError: On unexpected error copying file.
        """
        proc = self.execute_run(
            ["test", "-f", source.as_posix()],
            check=False,
            timeout=TIMEOUT_SIMPLE,
        )
        if proc.returncode != 0:
            raise FileNotFoundError(f"File not found: {source.as_posix()!r}")

        if not destination.parent.is_dir():
            raise FileNotFoundError(f"Directory not found: {str(destination.parent)!r}")

        self.lxc.file_pull(
            instance_name=self.instance_name,
            source=source,
            destination=destination,
            project=self.project,
            remote=self.remote,
        )

    def push_file(self, *, source: pathlib.Path, destination: pathlib.PurePath) -> None:
        """Copy a file from the host into the environment.

        The destination file is overwritten if it exists.

        :param source: Host file to copy.
        :param destination: Target environment file path to copy to.  Parent
            directory (destination.parent) must exist.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist.
        :raises LXDError: On unexpected error copying file.
        """
        if not source.is_file():
            raise FileNotFoundError(f"File not found: {str(source)!r}")

        proc = self.execute_run(
            ["test", "-d", destination.parent.as_posix()],
            check=False,
            timeout=TIMEOUT_SIMPLE,
        )
        if proc.returncode != 0:
            raise FileNotFoundError(
                f"Directory not found: {str(destination.parent.as_posix())!r}"
            )

        # Copy into target with uid/gid 0, rather than copying the IDs from the
        # host file.
        self.lxc.file_push(
            instance_name=self.instance_name,
            source=source,
            destination=destination,
            project=self.project,
            remote=self.remote,
            gid=0,
            uid=0,
        )

    def start(self) -> None:
        """Start instance.

        :raises LXDError: on unexpected error.
        """
        logger.info("Starting instance")
        self.lxc.start(
            instance_name=self.instance_name, project=self.project, remote=self.remote
        )

    def restart(self) -> None:
        """Restart instance.

        :raises LXDError: on unexpected error.
        """
        self.lxc.restart(
            instance_name=self.instance_name, project=self.project, remote=self.remote
        )

    def stop(self) -> None:
        """Stop instance.

        :raises LXDError: on unexpected error.
        """
        self.lxc.stop(
            instance_name=self.instance_name, project=self.project, remote=self.remote
        )

    def supports_mount(self) -> bool:
        """Check if instance supports mounting from host.

        :returns: True if mount is supported.
        """
        return self.remote == "local"

    def unmount(self, target: pathlib.PurePath) -> None:
        """Unmount mount target shared with host.

        :param target: Target shared with host to unmount.

        :raises LXDError: On failure to unmount target.
        """
        disks = self._get_disk_devices()

        unmounted = False
        for name, config in disks.items():
            if config["path"] == target.as_posix():
                self.lxc.config_device_remove(
                    instance_name=self.instance_name,
                    device=name,
                    project=self.project,
                    remote=self.remote,
                )
                unmounted = True

        if not unmounted:
            raise LXDError(
                brief=(
                    f"Failed to unmount {target.as_posix()!r}"
                    f" in instance {self.instance_name!r} - no such disk."
                ),
                details=f"* Disk device configuration: {disks!r}",
            )

    def unmount_all(self) -> None:
        """Unmount all mounts shared with host.

        :raises LXDError: On failure to unmount target.
        """
        disks = self._get_disk_devices()

        for name, _ in disks.items():
            self.lxc.config_device_remove(
                instance_name=self.instance_name,
                device=name,
                project=self.project,
                remote=self.remote,
            )

    def config_get(self, key: str) -> str:
        """Get instance configuration value.

        :param key: Configuration key to get.

        :returns: Configuration value.

        :raises LXDError: On unexpected error.
        """
        return self.lxc.config_get(
            instance_name=self.instance_name,
            key=key,
            project=self.project,
            remote=self.remote,
        )

    def config_set(self, key: str, value: str) -> None:
        """Set instance configuration value.

        :param key: Configuration key to set.
        :param value: Configuration key to the value.

        :returns: None.

        :raises LXDError: On unexpected error.
        """
        self.lxc.config_set(
            instance_name=self.instance_name,
            key=key,
            value=value,
            project=self.project,
            remote=self.remote,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import sys

import pytest
from craft_providers.bases.ubuntu import BuilddBase, BuilddBaseAlias
from craft_providers.multipass import MultipassProvider, is_installed


def test_ensure_provider_is_available_not_installed(uninstalled_multipass):
    """Verify multipass is installed and configured."""
    assert is_installed() is False
    provider = MultipassProvider()
    provider.ensure_provider_is_available()


def test_ensure_provider_is_available_installed(installed_multipass):
    """Verify multipass is installed and configured."""
    assert is_installed() is True
    provider = MultipassProvider()
    provider.ensure_provider_is_available()


def test_create_environment(installed_multipass, instance_name):
    """Verify create environment does not produce an error."""
    provider = MultipassProvider()
    test_instance = provider.create_environment(instance_name=instance_name)
    assert test_instance.exists() is False


ALIASES = list(BuilddBaseAlias)
ALIASES.remove(BuilddBaseAlias.XENIAL)


@pytest.mark.parametrize("alias", ALIASES)
def test_launched_environment(alias, installed_multipass, instance_name, tmp_path):
    """Verify `launched_environment()` creates and starts an instance then stops
    the instance when the method loses context."""
    if sys.platform == "darwin" and alias == BuilddBaseAlias.DEVEL:
        pytest.skip(reason="snapcraft:devel is not working on MacOS (LP #2007419)")

    project_path = tmp_path / "project"
    cache_path = tmp_path / "cache"

    provider = MultipassProvider()

    base_configuration = BuilddBase(alias=alias, cache_path=cache_path)

    with provider.launched_environment(
        project_name="test-multipass-project",
        project_path=project_path,
        base_configuration=base_configuration,
        instance_name=instance_name,
        allow_unstable=True,
    ) as test_instance:
        assert test_instance.exists() is True
        assert test_instance.is_running() is True

        test_instance.execute_run(["touch", "/root/.cache/pip/test-pip-cache"])

        assert (
            cache_path
            / base_configuration.compatibility_tag
            / str(base_configuration.alias)
            / "pip"
            / "test-pip-cache"
        ).exists()

    try:
        assert test_instance.exists() is True
        assert test_instance.is_running() is False
    finally:
        test_instance.delete()



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Fixtures for Multipass integration tests."""
import subprocess
import time
from contextlib import contextmanager

import pytest


@pytest.fixture(autouse=True, scope="session")
def installed_multipass_required(installed_multipass):
    """All Multipass integration tests required multipass to be installed."""


@contextmanager
def tmp_instance(
    *,
    instance_name: str,
    image_name: str = "22.04",
    cpus: str = "2",
    disk: str = "64G",
    mem: str = "1G",
):
    subprocess.run(
        [
            "multipass",
            "launch",
            image_name,
            "--name",
            instance_name,
            "--cpus",
            cpus,
            "--memory",
            mem,
            "--disk",
            disk,
        ],
        capture_output=True,
        check=True,
    )

    # Make sure container is ready
    for _ in range(0, 2400):
        proc = subprocess.run(
            [
                "multipass",
                "exec",
                instance_name,
                "--",
                "systemctl",
                "is-system-running",
            ],
            capture_output=True,
            check=False,
            text=True,
        )
        running_state = proc.stdout.strip()
        if running_state in ["running", "degraded"]:
            break
        time.sleep(0.1)

    yield instance_name

    # Cleanup if not purged by the test.
    subprocess.run(
        ["multipass", "delete", "--purge", instance_name],
        capture_output=True,
        check=False,
    )



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import pathlib
import subprocess

import pytest
from craft_providers.lxd import LXDError

from . import conftest


@pytest.fixture()
def instance(instance_name, session_project):
    with conftest.tmp_instance(
        name=instance_name, project=session_project
    ) as tmp_instance:
        yield tmp_instance


def test_exec(instance, lxc, session_project):
    proc = lxc.exec(
        instance_name=instance,
        command=["echo", "this is a test"],
        project=session_project,
        capture_output=True,
    )

    assert proc.stdout == b"this is a test\n"


def test_config_get_and_set(instance, instance_name, lxc, session_project):
    """Set and get config key/value pairs."""
    lxc.config_set(
        instance_name=instance,
        key="user.test-key",  # `user` namespace is for arbitrary config values
        value="test-value",
        project=session_project,
    )

    value = lxc.config_get(
        instance_name=instance, key="user.test-key", project=session_project
    )

    assert value == "test-value"


def test_config_get_non_existent_key(instance, instance_name, lxc, session_project):
    """Get a non-existent key and confirm the value is an empty string."""
    value = lxc.config_get(
        instance_name=instance, key="non-existant-key", project=session_project
    )

    assert not value


def test_copy(instance, instance_name, lxc, session_project):
    """Test `copy()` with default arguments."""
    destination_instance_name = instance_name + "-destination"

    # copy the instance to a new instance
    lxc.copy(
        source_instance_name=instance,
        destination_instance_name=destination_instance_name,
        project=session_project,
    )

    instances = lxc.list_names(project=session_project)

    # verify both instances exist
    assert instances == [instance, destination_instance_name]


def test_copy_error(instance, instance_name, lxc, session_project):
    """Raise a LXDError when the copy command fails."""
    # the source and destination cannot be same, so LXC will fail to copy
    with pytest.raises(LXDError) as raised:
        lxc.copy(
            source_instance_name=instance,
            destination_instance_name=instance,
            project=session_project,
        )

    assert raised.value == LXDError(
        brief=(
            f"Failed to copy instance 'local:{instance_name}' to 'local:"
            f"{instance_name}'."
        ),
        details=(
            f"* Command that failed: 'lxc --project {session_project} copy local:"
            f"{instance_name} local:{instance_name}'\n"
            "* Command exit code: 1\n"
            "* Command standard error output: b'Error: Failed creating instance "
            'record: Add instance info to the database: This "instances" entry '
            "already exists\\n'"
        ),
    )


def test_delete(instance, lxc, session_project):
    with pytest.raises(LXDError):
        lxc.delete(instance_name=instance, force=False, project=session_project)

    instances = lxc.list_names(project=session_project)
    assert instance in instances


def test_delete_force(instance, lxc, session_project):
    lxc.delete(instance_name=instance, force=True, project=session_project)

    instances = lxc.list_names(project=session_project)
    assert instance not in instances


def test_image_copy(lxc, session_project):
    lxc.image_copy(
        image="22.04",
        image_remote="ubuntu",
        alias="test-2204",
        project=session_project,
    )

    images = lxc.image_list(project=session_project)
    assert len(images) == 1


def test_image_delete(lxc, session_project):
    lxc.image_copy(
        image="22.04",
        image_remote="ubuntu",
        alias="test-2204",
        project=session_project,
    )

    lxc.image_delete(image="test-2204", project=session_project)

    images = lxc.image_list(project=session_project)
    assert images == []


def test_file_push(instance, lxc, session_project, tmp_path):
    test_file = tmp_path / "test.txt"
    test_file.write_text("this is a test")

    lxc.file_push(
        instance_name=instance,
        project=session_project,
        source=test_file,
        destination=pathlib.PurePosixPath("/tmp/foo"),
    )

    proc = lxc.exec(
        command=["cat", "/tmp/foo"],
        instance_name=instance,
        project=session_project,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=True,
    )

    assert proc.stdout == b"this is a test"


def test_file_pull(instance, lxc, session_project, tmp_path):
    out_path = tmp_path / "out.txt"
    test_file = tmp_path / "test.txt"
    test_file.write_text("this is a test")

    lxc.file_push(
        instance_name=instance,
        project=session_project,
        source=test_file,
        destination=pathlib.PurePosixPath("/tmp/foo"),
    )

    lxc.file_pull(
        instance_name=instance,
        project=session_project,
        source=pathlib.PurePosixPath("/tmp/foo"),
        destination=out_path,
    )

    assert out_path.read_text() == "this is a test"


def test_disk_add_remove(instance, lxc, tmp_path, session_project):
    mount_target = pathlib.PurePosixPath("/mnt")

    # Make sure permissions allow read from inside guest without mappings.
    tmp_path.chmod(0o755)

    test_file = tmp_path / "test.txt"
    test_file.write_text("this is a test")

    lxc.config_device_add_disk(
        instance_name=instance,
        source=tmp_path,
        path=mount_target,
        device="test_mount",
        project=session_project,
    )

    proc = lxc.exec(
        command=["cat", "/mnt/test.txt"],
        instance_name=instance,
        capture_output=True,
        check=True,
        project=session_project,
    )

    assert proc.stdout == b"this is a test"

    lxc.config_device_remove(
        instance_name=instance,
        device="test_mount",
        project=session_project,
    )

    with pytest.raises(subprocess.CalledProcessError):
        lxc.exec(
            command=["test", "-f", "/mnt/test.txt"],
            project=session_project,
            instance_name=instance,
            check=True,
        )


def test_info(instance, lxc, session_project):
    """Test `info()` method works as expected."""
    data = lxc.info(instance_name=instance, project=session_project)

    assert data["Name"] == instance
